# coding: utf-8
#
# dict.py
#
# Copyright (C) 2020 IMTEK Simulation
# Author: Johannes Hoermann, johannes.hoermann@imtek.uni-freiburg.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Utility functions for nested dicts and lists."""

import collections
import logging

from fireworks.utilities.dict_mods import get_nested_dict_value, apply_mod
from jlhfw.utils.logging import _log_nested_dict

__author__ = 'Johannes Laurin Hoermann'
__copyright__ = 'Copyright 2020, IMTEK Simulation, University of Freiburg'
__email__ = 'johannes.hoermann@imtek.uni-freiburg.de, johannes.laurin@gmail.com'
__date__ = 'August 27, 2020'


# from https://gist.github.com/angstwad/bf22d1822c38a92ec0a9
def simple_dict_merge(dct, merge_dct, add_keys=True):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    This version will return a copy of the dictionary and leave the original
    arguments untouched.

    The optional argument ``add_keys``, determines whether keys which are
    present in ``merge_dict`` but not ``dct`` should be included in the
    new dict.

    Args:
        dct (dict) onto which the merge is executed
        merge_dct (dict): dct merged into dct
        add_keys (bool): whether to add new keys

    Returns:
        dict: updated dict
    """
    dct = dct.copy()
    if not add_keys:
        merge_dct = {
            k: merge_dct[k]
            for k in set(dct).intersection(set(merge_dct))
        }

    for k, v in merge_dct.items():
        if (k in dct and isinstance(dct[k], dict)
                and isinstance(v, collections.Mapping)):
            dct[k] = dict_merge(dct[k], v, add_keys=add_keys)
        else:
            dct[k] = v

    return dct


# from https://gist.github.com/angstwad/bf22d1822c38a92ec0a9
def dict_merge(dct, merge_dct, add_keys=True,
               exclusions={}, merge_exclusions={}):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    This version will return a copy of the dictionary and leave the original
    arguments untouched.

    The optional argument ``add_keys``, determines whether keys which are
    present in ``merge_dct`` but not ``dct`` should be included in the
    new dict.

    Args:
        dct (dict) onto which the merge is executed
        merge_dct (dict): dct merged into dct
        add_keys (bool): whether to add new keys
        exclusions (dict): any such key found within 'dct' will be removed.
                               It can, however, be reintroduced if present in
                               'merge_dct' and 'add_keys' is set.
        merge_exclusions (dict): any such key found within 'merge_dct'
                                     will be ignored. Such keys allready
                                     present within 'dct' are not touched.

    Returns:
        dict: updated dict
    """
    logger = logging.getLogger(__name__)

    merge_dct = merge_dct.copy()
    dct = dct.copy()

    logger.debug("Merge 'merge_dct'...")
    _log_nested_dict(logger.debug, merge_dct)
    logger.debug("... into 'dct'...")
    _log_nested_dict(logger.debug, dct)
    logger.debug("... with 'exclusions'...")
    _log_nested_dict(logger.debug, exclusions)
    logger.debug("... and 'merge_exclusions'...")
    _log_nested_dict(logger.debug, merge_exclusions)

    # remove all entries marked as exclusions from dct
    for k in exclusions:
        if (k in dct) and (exclusions[k] is True):
            logger.debug("Key '{}' excluded from dct.".format(k))
            del dct[k]

    if not add_keys:
        merge_dct = {
            k: merge_dct[k] for k in set(dct).intersection(set(merge_dct))}
        logger.debug(
            "Not merging keys only in 'merge_dict', only merging {}.".format(
                merge_dct.keys()))

    # Add empty fields to merge_dct for all keys in dct but not in merge_dct.
    # This makes sure the union of all keys from dct stripped off exclusions
    # and merge_dct not yet stripped off merge_exclusions is treated in the
    # following.
    for k in dct.keys():
        if isinstance(dct[k], dict) and k not in merge_dct:
            merge_dct[k] = {}

    for k, v in merge_dct.items():
        # all truly excluded keys allready removed from dct, thus an existing
        # entry in exclusions means that some more specific nested key
        # is excluded
        if k in exclusions:
            lower_level_exclusions = exclusions[k]
            logger.debug("Key '{}' included in dct, but exclusions exist for nested keys.".format(k))
        else:
            lower_level_exclusions = {}
            logger.debug("Key '{}' included in dct.".format(k))

        # only treat nested nested dicts for now
        # TODO: treat also nested lists
        if (k in dct and isinstance(dct[k], dict)
                and isinstance(v, collections.Mapping)):
            if k not in merge_exclusions:  # no exception rule for this field
                logger.debug("Key '{}' included in merge.".format(k))
                dct[k] = dict_merge(dct[k], v, add_keys=add_keys,
                                    exclusions=lower_level_exclusions)
            elif merge_exclusions[k] is not True:  # exclusion rule for nested fields
                logger.debug("Key '{}' included in merge, but exclusions exist for nested keys.".format(k))
                dct[k] = dict_merge(dct[k], v, add_keys=add_keys,
                                    exclusions=lower_level_exclusions,
                                    merge_exclusions=merge_exclusions[k])
            else:
                logger.debug("Key '{}' excluded from merge.".format(k))
        else:
            if k not in merge_exclusions:  # no exception rule for this field
                logger.debug("Key '{}' included in merge.".format(k))
                dct[k] = v
            else:
                logger.debug("Key '{}' excluded from merge.".format(k))

    return dct


def _make_marker(d):
    """Mark everything for comparison."""
    if isinstance(d, list):
        return [_make_marker(e) for e in d]
    elif isinstance(d, dict):
        return {k: _make_marker(v) for k, v in d.items()}
    else:
        return True

def _compare(source, target, marker):
    """Compare source and target partially, as marked by marker."""
    logger = logging.getLogger(__name__)
    if isinstance(marker, dict):
        for k, v in marker.items():
            if k not in source:
                logger.error("{} not in source '{}'.".format(k, source))
                return False
            if k not in target:
                logger.error("{} not in target '{}'.".format(k, source))
                return False

            logger.debug("Descending into sub-tree '{}' of '{}'.".format(
                source[k], source))
            # descend
            if not _compare(source[k], target[k], v):
                return False  # one failed comparison suffices

    elif isinstance(marker, list):  # source, target and marker must have same length
        logger.debug("Branching into element wise sub-trees of '{}'.".format(
            source))
        for s, t, m in zip(source, target, marker):
            if not _compare(s, t, m):
                return False  # one failed comparison suffices
    else:  # arrived at leaf, comparison desired?
        if marker is not False:  # yes
            logger.debug("Comparing '{}' == '{}' -> {}.".format(
                source, target, source == target))
            return source == target

    # comparison either not desired or successfull for all elements
    return True


def compare(source, target, marker=None):
    """Compare source and target partially, as marked by marker. If marker is None, then compare everything."""
    if not marker:
        marker = _make_marker(source)
    return _compare(source, target, marker)


def from_fw_spec(param, fw_spec):
    """Expands param['key'] as key within fw_spec.

    If param is dict hand has field 'key', then return value at specified
    position from fw_spec. Otherwise, return 'param' itself.
    """
    if isinstance(param, dict) and 'key' in param:
        ret = get_nested_dict_value(fw_spec, param['key'])
    else:
        ret = param
    return ret


# we apply update_spec and mod_spec here a priori because additions and detours
# won't be touched by the default mechanism in fireworks.core.firework
def apply_mod_spec(wf, action, fw_ids=None):
    """Update the spec of the children FireWorks using DictMod language."""
    fw_ids = fw_ids if fw_ids else wf.leaf_fw_ids
    updated_ids = []

    if action.update_spec and action.propagate:
        # Traverse whole sub-workflow down to leaves.
        visited_cfid = set()  # avoid double-updating for diamond deps

        def recursive_update_spec(fw_ids):
            for cfid in fw_ids:
                if cfid not in visited_cfid:
                    visited_cfid.add(cfid)
                    wf.id_fw[cfid].spec.update(action.update_spec)
                    updated_ids.append(cfid)
                    recursive_update_spec(wf.links[cfid])

        recursive_update_spec(fw_ids)
    elif action.update_spec:
        # Update only direct children.
        for cfid in fw_ids:
            wf.id_fw[cfid].spec.update(action.update_spec)
            updated_ids.append(cfid)

    if action.mod_spec and action.propagate:
        visited_cfid = set()

        def recursive_mod_spec(fw_ids):
            for cfid in fw_ids:
                if cfid not in visited_cfid:
                    visited_cfid.add(cfid)
                    for mod in action.mod_spec:
                        apply_mod(mod, wf.id_fw[cfid].spec)
                    updated_ids.append(cfid)
                    recursive_mod_spec(cfid)

        recursive_mod_spec(fw_ids)
    elif action.mod_spec:
        for cfid in fw_ids:
            for mod in action.mod_spec:
                apply_mod(mod, wf.id_fw[cfid].spec)
            updated_ids.append(cfid)

    return updated_ids



# coding: utf-8
"""Fixtures."""
__author__ = 'Johannes Laurin Hoermann'
__copyright__ = 'Copyright 2020, IMTEK Simulation, University of Freiburg'
__email__ = 'johannes.hoermann@imtek.uni-freiburg.de, johannes.laurin@gmail.com'
__date__ = 'Nov 6, 2020'

import json
import logging
import os
import pytest
import tempfile

from jlhfw.utils.logging import _log_nested_dict

module_dir = os.path.abspath(os.path.dirname(__file__))


def _read_json(file):
    with open(file, 'r') as stream:
        return json.load(stream)


@pytest.fixture
def files():
    """Provide paths to dtool tasks tests files."""
    return {
        'dtool_readme_static_and_dynamic_metadata_test':
            os.path.join(
                module_dir,
                "dtool_readme_static_and_dynamic_metadata_test.yml"),
        'dtool_readme_dynamic_metadata_test':
            os.path.join(
                module_dir, "dtool_readme_dynamic_metadata_test.yml"),
        'dtool_readme_static_metadata_test':
            os.path.join(
                module_dir, "dtool_readme_static_metadata_test.yml"),
        'dtool_readme_template_path':
            os.path.join(module_dir, "dtool_readme.yml"),
        'dtool_config_path':
            os.path.join(module_dir, "dtool.json"),
    }


@pytest.fixture
def dtool_config(files):
    """Provide default dtool config."""
    logger = logging.getLogger(__name__)
    # use environment variables instead of custom config file, see
    # https://github.com/jic-dtool/dtoolcore/pull/17
    # _original_environ = os.environ.copy()

    # inject configuration into environment:
    dtool_config = _read_json(files['dtool_config_path'])
    logger.debug("dtool config overrides:")
    _log_nested_dict(logger.debug, dtool_config)

    return dtool_config


@pytest.fixture
def tempdir(request):
    """Provide clean temporary directory."""
    tmpdir = tempfile.TemporaryDirectory()
    previous_working_directory = os.getcwd()
    os.chdir(tmpdir.name)

    def finalizer():
        os.chdir(previous_working_directory)
        tmpdir.cleanup()

    request.addfinalizer(finalizer)
    return



# coding: utf-8
#
# logging.py
#
# Copyright (C) 2020 IMTEK Simulation
# Author: Johannes Hoermann, johannes.hoermann@imtek.uni-freiburg.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Logging utils."""

import json
import logging


def _log_nested_dict(log_func, dct):
    for l in json.dumps(dct, indent=2, default=str).splitlines():
        log_func(l)

class LoggingContext():
    """Modifies logging within context."""
    def __init__(self, logger=None, handler=None, level=None, close=False):
        self.logger = logger
        self.level = level
        self.handler = handler
        self.close = close

    def __enter__(self):
        """Prepare log output streams."""
        if self.logger is None:
            # temporarily modify root logger
            self.logger = logging.getLogger('')
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)
        if self.handler:
            self.logger.addHandler(self.handler)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
        if self.handler:
            self.logger.removeHandler(self.handler)
        if self.handler and self.close:
            self.handler.close()
        # implicit return of None => don't swallow exceptions



# coding: utf-8
#
# recover_tasks.py
#
# Copyright (C) 2020 IMTEK Simulation
# Author: Johannes Hoermann, johannes.hoermann@imtek.uni-freiburg.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Tasks that recover and restart failed computations from checkpoints."""

import glob
import io
import logging
import os
import shutil

from collections.abc import Iterable
from contextlib import ExitStack

from fireworks.fw_config import FW_LOGGING_FORMAT
from fireworks.utilities.fw_serializers import ENCODING_PARAMS
from fireworks.utilities.dict_mods import get_nested_dict_value, set_nested_dict_value
from fireworks.core.firework import FWAction, FiretaskBase, Firework, Workflow

from jlhfw.utils.logging import LoggingContext, _log_nested_dict
from jlhfw.utils.dict import dict_merge, from_fw_spec, apply_mod_spec


__author__ = 'Johannes Laurin Hoermann'
__copyright__ = 'Copyright 2020, IMTEK Simulation, University of Freiburg'
__email__ = 'johannes.hoermann@imtek.uni-freiburg.de, johannes.laurin@gmail.com'
__date__ = 'August 10, 2020'

DEFAULT_FORMATTER = logging.Formatter(FW_LOGGING_FORMAT)

#TODO: if wrong parent fails (i.e. unexpected order of entries in _fizzled_parents) then wrong launchdir is
# selected to copy everything (including restart files from). Correct i.e. by matching against some identifier.

class RecoverTask(FiretaskBase):
    """
    Generic base class for recovering and restarting some computation.

    The typical use case is some FIZZLED run, i.e. due to exceeded walltime.
    Activate _allow_fizzled_parents and append to initial run Firework.

    If inserted by means of some 'recovery fw' between a parent and its children

        +--------+     +------------+
        | parent | --> | child(ren) |
        +--------+     +------------+

    as shown

        +--------+     +-------------+     +------------+
        | parent | --> | recovery fw | --> | child(ren) |
        +--------+     +-------------+     +------------+


    then this task generates the following insertion in case of the parent's
    failure

                                                       +- - - - - - - - - - - - - - - -+
                                                       ' detour_wf                     '
                                                       '                               '
                                                       ' +-----------+     +---------+ '
                                 +-------------------> ' |  root(s)  | --> | leaf(s) | ' ------+
                                 |                     ' +-----------+     +---------+ '       |
                                 |                     '                               '       |
                                 |                     +- - - - - - - - - - - - - - - -+       |
                                 |                                                             |
                                 |                                                             |
                                 |                                                             |
                                                       +- - - - - - - - - - - - - - - -+       |
                                                       ' restart_wf                    '       |
                                                       '                               '       v
        +----------------+     +-----------------+     ' +-----------+     +---------+ '     +-----------------+     +------------+
        | fizzled parent | --> | 1st recovery fw | --> ' |  root(s)  | --> | leaf(s) | ' --> | 2nd recovery fw | --> | child(ren) |
        +----------------+     +-----------------+     ' +-----------+     +---------+ '     +-----------------+     +------------+
                                                       '                               '
                                                       +- - - - - - - - - - - - - - - -+
                                 |
                                 |
                                 |
                                 |                     +- - - - - - - - - - - - - - - -+
                                 |                     ' addition_wf                   '
                                 |                     '                               '
                                 |                     ' +-----------+     +---------+ '
                                 +-------------------> ' |  root(s)  | --> | leaf(s) | '
                                                       ' +-----------+     +---------+ '
                                                       '                               '
                                                       +- - - - - - - - - - - - - - - -+

    This dynamoc insertion repeats until 'restart_wf' completes successfully
    or the number of repetitions reaches 'max_restarts'. While 'restart_wf'
    is only appended in the case of a parent's failure, 'detour_wf' and
    'addition_wf' are always inserted:

        + - - - - - - - - - - - -+                              + - - - - - - - - - - - - - - - +
        ' successfull restart_wf '                              ' detour_wf                     '
        '                        '                              '                               '
        '            +---------+ '     +------------------+     ' +-----------+     +---------+ '     +------------+
        '   ...  --> | leaf(s) | ' --> | last recovery fw | --> ' |  root(s)  | --> | leaf(s) | ' --> | child(ren) |
        '            +---------+ '     +------------------+     ' +-----------+     +---------+ '     +------------+
        '                        '                              '                               '
        + - - - - - - - - - - - -+                              + - - - - - - - - - - - - - - - +
                                                           |
                                                           |
                                                           |
                                                           |                      + - - - - - - - - - - - - - - - +
                                                           |                      ' addition_wf                   '
                                                           |                      '                               '
                                                           |                      ' +-----------+     +---------+ '
                                                           +--------------------> ' |  root(s)  | --> | leaf(s) | '
                                                                                  ' +-----------+     +---------+ '
                                                                                  '                               '
                                                                                  + - - - - - - - - - - - - - - - +

    NOTE: make sure that the used 'recovery fw' forwards all outputs
    transparently in case of parent's success.

    NOTE: while the dynamic extensions 'detour_wf' and 'addition_wf' can
    actually be a whole Workflow as well as a single FireWork, 'recover_fw'
    must be single FireWorks. If more complex constructs are necessary,
    consider generating such within those FireWorks.

    NOTE: fails for several parents if the "wrong" parent fizzles. Use only
    in unambiguous situations.

    Required parameters:
        - restart_wf (dict): Workflow or single FireWork to append only if
            restart file present (parent failed). Task will not append anything
            if None. Default: None

    Optional parameters:
        - addition_wf (dict): Workflow or single FireWork to always append as
            an addition, independent on parent's success (i.e. storage).
            Default: None.
        - detour_wf (dict): Workflow or single FireWork to always append as
            a detour, independent on parent's success (i.e. post-processing).
            Task will not append anything if None. Default: None

        - restart_fws_root ([int]): fw_ids (referring to fws in restart_wf)
            to identify "roots" to be connected to this recover_fw. If not
            specified, all dangling roots are connected.
        - restart_fws_leaf ([int]): fw_ids (referring to fws in restart_wf)
            to identify "leaves" to be connected to next recover_fw. If not
            specified, all dangling leaves are connected.
        - detour_fws_root ([int]): fw_ids (referring to fws in detour_wf)
            to identify "roots" to be connected to this recover_fw. If not
            specified, all dangling roots are connected.
        - detour_fws_leaf ([int]): fw_ids (referring to fws in detour_wf)
            to identify "leaves" to be connected to next recover_fw. If not
            specified, all dangling leaves are connected.
        - addition_fws_root ([int]): fw_ids (referring to fws in addition_wf)
            to identify "roots" to be connected to this recover_fw. If not
            specified, all dangling roots are connected.

        - apply_mod_spec_to_addition_wf (bool): Apply FWAction's update_spec and
            mod_spec to 'addition_wf', same as for all other regular childern
            of this task's FireWork. Default: True.
        - apply_mod_spec_to_detour_wf (bool): Apply FWAction's update_spec and
            mod_spec to 'detour_wf', 'restart_wf', and the next instance of
            'recover_fw', same as for all other regular children of this task's
            FireWork. Default: True.
        - fizzle_on_no_restart_file (bool): Default: True
        - fw_spec_to_exclude ([str]): recover FireWork will inherit the current
            FireWork's 'fw_spec', stripped of top-level fields specified here.
            Default: ['_job_info', '_fw_env', '_files_prev', '_fizzled_parents']
        - ignore_errors (bool): Ignore errors when copying files. Default: True
        - max_restarts (int): Maximum number of repeated restarts (in case of
            'restart' == True). Default: 5
        - other_glob_patterns (str or [str]): Patterns for glob.glob to identify
            other files to be forwarded. All files matching this glob pattern
            are recovered. Default: None
        - repeated_recover_fw_name (str): Name for repeated recovery fireworks.
            If None, the name of this FireWorksis used. Default: None.
        - restart_counter (str): fw_spec path for restart counting.
            Default: 'restart_count'.
        - restart_file_glob_patterns (str or [str]): Patterns for glob.glob to
            identify restart files. Attention: Be careful not to match any
            restart file that has been used as an input initially.
            If more than one file matches a glob pattern in the list, only the
            most recent macth per list entry is recovered.
            Default: ['*.restart[0-9]']
        - restart_file_dests (str or [str]): Destinations of copied restart
            files. Most recent restart file found for each entry in the
            restart_file_glob_patterns list will be copied to new launchdir
            named after the according entry within this list. If None, then
            files keep their original names. Default: None
        - superpose_restart_on_parent_fw_spec (bool):
            Try to pull (fizzled) parent's fw_spec and merge with fw_spec of
            all FireWorks within restart_wf, with latter enjoying precedence.
            Default: False.
        - superpose_addition_on_parent_fw_spec (bool):
            Try to pull (fizzled) parent's fw_spec and merge with fw_spec of
            all FireWorks within addition_wf, with latter enjoying precedence.
            Default: False.
        - superpose_detour_on_parent_fw_spec (bool):
            Try to pull (fizzled) parent's fw_spec and merge with fw_spec of
            all FireWorks within detour_wf, with latter enjoying precedence.
            Default: False.
        - restart_fw_spec_to_exclude ([str]):
        - addition_fw_spec_to_exclude ([str]):
        - detour_fw_spec_to_exclude ([str]):
            When any of the above superpose flags is set, the top-level fw_spec
            fields specified here won't enter the created additions.
            Default for all three are all reserved fields, i.e. [
                '_add_fworker',
                '_add_launchpad_and_fw_id',
                '_allow_fizzled_parents',
                '_background_tasks',
                '_category',
                '_dupefinder',
                '_files_in',
                '_files_out',
                '_files_prev',
                '_fizzled_parents',
                '_fw_env',
                '_fworker',
                '_job_info',
                '_launch_dir',
                '_pass_job_info',
                '_preserve_fworker',
                '_priority',
                '_queueadapter',
                '_tasks',
                '_trackers',
            ]

        - output (str): spec key that will be used to pass output to child
            fireworks. Default: None
        - dict_mod (str, default: '_set'): how to insert output into output
            key, see fireworks.utils.dict_mods
        - propagate (bool, default: None): if True, then set the
            FWAction 'propagate' flag and propagate updated fw_spec not only to
            direct children, but to all descendants down to wokflow's leaves.
        - stored_data (bool, default: False): put outputs into database via
            FWAction.stored_data
        - store_stdlog (bool, default: False): insert log output into database
            (only if 'stored_data' or 'output' is spcified)
        - stdlog_file (str, Default: NameOfTaskClass.log): print log to file
        - loglevel (str, Default: logging.INFO): loglevel for this task

    Fields 'max_restarts', 'restart_file_glob_patterns', 'other_glob_patterns',
    'default_restart_file', 'fizzle_on_no_restart_file',
    'repeated_recover_fw_name', and 'ignore_errors'
    may also be a dict of format { 'key': 'some->nested->fw_spec->key' } for
    looking up value within 'fw_spec' instead.

    NOTE: reserved fw_spec keywords are (alphabetically)
        - _add_fworker
        - _add_launchpad_and_fw_id
        - _allow_fizzled_parents
        - _background_tasks
        - _category
        - _dupefinder
        - _files_in
        - _files_out
        - _files_prev
        - _fizzled_parents
        - _fw_env
        - _fworker
        - _job_info
        - _launch_dir
        - _pass_job_info
        - _preserve_fworker
        - _priority
        - _queueadapter
        - _tasks
        - _trackers
    """
    _fw_name = "RecoverTask"
    required_params = [
        "restart_wf",
    ]
    optional_params = [
        "detour_wf",
        "addition_wf",

        "restart_fws_root",
        "restart_fws_leaf",
        "detour_fws_root",
        "detour_fws_leaf",
        "addition_fws_root",

        "apply_mod_spec_to_addition_wf",
        "apply_mod_spec_to_detour_wf",
        "fizzle_on_no_restart_file",
        "fw_spec_to_exclude",
        "ignore_errors",
        "max_restarts",
        "other_glob_patterns",
        "repeated_recover_fw_name",
        "restart_counter",
        "restart_file_glob_patterns",
        "restart_file_dests",

        "superpose_restart_on_parent_fw_spec",
        "superpose_addition_on_parent_fw_spec",
        "superpose_detour_on_parent_fw_spec",
        "restart_fw_spec_to_exclude",
        "addition_fw_spec_to_exclude",
        "detour_fw_spec_to_exclude",

        "stored_data",
        "output",
        "dict_mod",
        "propagate",
        "stdlog_file",
        "store_stdlog",
        "loglevel"]

    def appendable_wf_from_dict(self, obj_dict, base_spec=None, exclusions={}):
        """Creates Workflow from a Workflow or single FireWork dict description.

        If specified, use base_spec for all fw_spec and superpose individual
        specs on top.

        Args:
            - obj_dict (dict): describes either single FW or whole Workflow
            - base_spec (dict): use those specs for all FWs within workflow.
                Specific specs already set within obj_dict take precedence.
            - exclusions (dict): nested dict with keys marked for exclusion
                by True boolean value. Excluded keys are stripped off base_spec.

        Returns:
            (Workflow, dict) tuple
            - Worfklow: created workflow
            - {int: int}: old to new fw_ids mapping
        """
        logger = logging.getLogger(__name__)

        logger.debug("Initial obj_dict:")
        _log_nested_dict(logger.debug, obj_dict)

        if base_spec:
            logger.debug("base_spec:")
            _log_nested_dict(logger.debug, base_spec)

        if exclusions:
            logger.debug("exclusions:")
            _log_nested_dict(logger.debug, exclusions)

        remapped_fw_ids = {}
        if isinstance(obj_dict, dict):
            # in case of single Fireworks:
            if "spec" in obj_dict:
                # append firework (defined as dict):
                if base_spec:
                    obj_dict["spec"] = dict_merge(base_spec, obj_dict["spec"],
                                                  exclusions=exclusions)
                fw = Firework.from_dict(obj_dict)
                remapped_fw_ids[fw.fw_id] = self.consecutive_fw_id
                fw.fw_id = self.consecutive_fw_id
                self.consecutive_fw_id -= 1
                wf = Workflow([fw])
            else:   # if no single fw, then wf
                if base_spec:
                    for fw_dict in obj_dict["fws"]:
                        fw_dict["spec"] = dict_merge(base_spec, fw_dict["spec"],
                                                     exclusions=exclusions)
                wf = Workflow.from_dict(obj_dict)
                # do we have to reassign fw_ids? yes
                for fw in wf.fws:
                    remapped_fw_ids[fw.fw_id] = self.consecutive_fw_id
                    fw.fw_id = self.consecutive_fw_id
                    self.consecutive_fw_id -= 1
                wf._reassign_ids(remapped_fw_ids)
        else:
            raise ValueError("type({}) is '{}', but 'dict' expected.".format(
                             obj_dict, type(obj_dict)))
        logger.debug("Built object:")
        _log_nested_dict(logger.debug, wf.as_dict())

        return wf, remapped_fw_ids

    # modeled to match original snippet from fireworks.core.rocket:
    #
    # if my_spec.get("_files_out"):
    #     # One potential area of conflict is if a fw depends on two fws
    #     # and both fws generate the exact same file. That can lead to
    #     # overriding. But as far as I know, this is an illogical use
    #     # of a workflow, so I can't see it happening in normal use.
    #     for k, v in my_spec.get("_files_out").items():
    #         files = glob.glob(os.path.join(launch_dir, v))
    #         if files:
    #             filepath = sorted(files)[-1]
    #             fwaction.mod_spec.append({
    #                 "_set": {"_files_prev->{:s}".format(k): filepath}
    #             })

    # if the curret fw yields outfiles, then check whether according
    # '_files_prev' must be written for newly created insertions
    def write_files_prev(self, wf, fw_spec, root_fw_ids=None):
        "Sets _files_prev in roots of new workflow according to _files_out in fw_spec."
        logger = logging.getLogger(__name__)

        if fw_spec.get("_files_out"):
            logger.info("Current FireWork's '_files_out': {}".format(
                        fw_spec.get("_files_out")))

            files_prev = {}

            for k, v in fw_spec.get("_files_out").items():
                files = glob.glob(os.path.join(os.curdir, v))
                if files:
                    logger.info("This Firework provides {}: {}".format(
                                k, files), " within _files_out.")
                    filepath = os.path.abspath(sorted(files)[-1])
                    logger.info("{}: '{}' provided as '_files_prev'".format(
                                k, filepath), " to subsequent FireWorks.")
                    files_prev[k] = filepath

            # get roots of insertion wf and assign _files_prev to them
            if root_fw_ids is None:
                root_fw_ids = wf.root_fw_ids
            root_fws = [fw for fw in wf.fws if fw.fw_id in root_fw_ids]

            for root_fw in root_fws:
                root_fw.spec["_files_prev"] = files_prev

        return wf

    def run_task(self, fw_spec):
        # NOTE: be careful to distinguish between what is referred to as
        # detour_wf in the task's docstring, stored within detour_wf_dict below
        # and the final detour_wf constructed, possibly comprising the partial
        # restart and detour workflows specified via task parameters as well as
        # another copy of this recover_fw.

        self.consecutive_fw_id = -1  # quite an ugly necessity
        # get fw_spec entries or their default values:
        restart_wf_dict = self.get('restart_wf', None)
        detour_wf_dict = self.get('detour_wf', None)
        addition_wf_dict = self.get('addition_wf', None)

        restart_fws_root = self.get('restart_fws_root', None)
        restart_fws_leaf = self.get('restart_fws_leaf', None)
        detour_fws_root = self.get('detour_fws_root', None)
        detour_fws_leaf = self.get('detour_fws_leaf', None)
        addition_fws_root = self.get('addition_fws_root', None)

        apply_mod_spec_to_addition_wf = self.get('apply_mod_spec_to_addition_wf', True)
        apply_mod_spec_to_addition_wf = from_fw_spec(apply_mod_spec_to_addition_wf,
                                                     fw_spec)

        apply_mod_spec_to_detour_wf = self.get('apply_mod_spec_to_detour_wf', True)
        apply_mod_spec_to_detour_wf = from_fw_spec(apply_mod_spec_to_detour_wf,
                                                   fw_spec)

        fizzle_on_no_restart_file = self.get('fizzle_on_no_restart_file', True)
        fizzle_on_no_restart_file = from_fw_spec(fizzle_on_no_restart_file,
                                                 fw_spec)

        ignore_errors = self.get('ignore_errors', True)
        ignore_errors = from_fw_spec(ignore_errors, fw_spec)

        max_restarts = self.get('max_restarts', 5)
        max_restarts = from_fw_spec(max_restarts, fw_spec)

        other_glob_patterns = self.get('other_glob_patterns', None)
        other_glob_patterns = from_fw_spec(other_glob_patterns, fw_spec)

        repeated_recover_fw_name = self.get('repeated_recover_fw_name',
                                            'Repeated LAMMPS recovery')
        repeated_recover_fw_name = from_fw_spec(repeated_recover_fw_name,
                                                fw_spec)

        restart_counter = self.get('restart_counter', 'restart_count')

        restart_file_glob_patterns = self.get('restart_file_glob_patterns',
                                              ['*.restart[0-9]'])
        restart_file_glob_patterns = from_fw_spec(restart_file_glob_patterns,
                                                  fw_spec)

        restart_file_dests = self.get('restart_file_dests', None)
        restart_file_dests = from_fw_spec(restart_file_dests, fw_spec)

        superpose_restart_on_parent_fw_spec = self.get(
            'superpose_restart_on_parent_fw_spec', False)
        superpose_restart_on_parent_fw_spec = from_fw_spec(
            superpose_restart_on_parent_fw_spec, fw_spec)

        superpose_addition_on_parent_fw_spec = self.get(
            'superpose_addition_on_parent_fw_spec', False)
        superpose_addition_on_parent_fw_spec = from_fw_spec(
            superpose_addition_on_parent_fw_spec, fw_spec)

        superpose_detour_on_parent_fw_spec = self.get(
            'superpose_detour_on_parent_fw_spec', False)
        superpose_detour_on_parent_fw_spec = from_fw_spec(
            superpose_detour_on_parent_fw_spec, fw_spec)

        fw_spec_to_exclude = self.get('fw_spec_to_exclude',
                                      [
                                        '_job_info',
                                        '_fw_env',
                                        '_files_prev',
                                        '_fizzled_parents',
                                      ])
        if isinstance(fw_spec_to_exclude, list):
            fw_spec_to_exclude_dict = {k: True for k in fw_spec_to_exclude}
        else:  # supposed to be dict then
            fw_spec_to_exclude_dict = fw_spec_to_exclude

        default_fw_spec_to_exclude = [
            '_add_fworker',
            '_add_launchpad_and_fw_id',
            '_allow_fizzled_parents',
            '_background_tasks',
            '_category',
            '_dupefinder',
            '_files_in',
            '_files_out',
            '_files_prev',
            '_fizzled_parents',
            '_fw_env',
            '_fworker',
            '_job_info',
            '_launch_dir',
            '_pass_job_info',
            '_preserve_fworker',
            '_priority',
            '_queueadapter',
            '_tasks',
            '_trackers',
        ]
        restart_fw_spec_to_exclude = self.get('restart_fw_spec_to_exclude', default_fw_spec_to_exclude)
        if isinstance(restart_fw_spec_to_exclude, list):
            restart_fw_spec_to_exclude_dict = {k: True for k in restart_fw_spec_to_exclude}
        else:  # supposed to be dict then
            restart_fw_spec_to_exclude_dict = restart_fw_spec_to_exclude

        addition_fw_spec_to_exclude = self.get('addition_fw_spec_to_exclude', default_fw_spec_to_exclude)
        if isinstance(addition_fw_spec_to_exclude, list):
            addition_fw_spec_to_exclude_dict = {k: True for k in addition_fw_spec_to_exclude}
        else:  # supposed to be dict then
            addition_fw_spec_to_exclude_dict = addition_fw_spec_to_exclude

        detour_fw_spec_to_exclude = self.get('detour_fw_spec_to_exclude', default_fw_spec_to_exclude)
        if isinstance(detour_fw_spec_to_exclude, list):
            detour_fw_spec_to_exclude_dict = {k: True for k in detour_fw_spec_to_exclude}
        else:  # supposed to be dict then
            detour_fw_spec_to_exclude_dict = detour_fw_spec_to_exclude

        # generic parameters
        stored_data = self.get('stored_data', False)
        output_key = self.get('output', None)
        dict_mod = self.get('dict_mod', '_set')
        propagate = self.get('propagate', False)

        stdlog_file = self.get('stdlog_file', '{}.log'.format(self._fw_name))
        store_stdlog = self.get('store_stdlog', False)
        loglevel = self.get('loglevel', logging.INFO)

        with ExitStack() as stack:

            if store_stdlog:
                stdlog_stream = io.StringIO()
                logh = logging.StreamHandler(stdlog_stream)
                logh.setFormatter(DEFAULT_FORMATTER)
                stack.enter_context(
                    LoggingContext(handler=logh, level=loglevel, close=False))

            # logging to dedicated log file if desired
            if stdlog_file:
                logfh = logging.FileHandler(
                    stdlog_file, mode='a', **ENCODING_PARAMS)
                logfh.setFormatter(DEFAULT_FORMATTER)
                stack.enter_context(
                    LoggingContext(handler=logfh, level=loglevel, close=True))

            logger = logging.getLogger(__name__)

            # input assertions, ATTENTION: order matters

            # avoid iterating through each character of string
            if isinstance(restart_file_glob_patterns, str):
                restart_file_glob_patterns = [restart_file_glob_patterns]

            if not restart_file_dests:
                # don't rename restart files when recovering
                restart_file_dests = os.curdir

            if isinstance(restart_file_dests, str):
                # if specified as plain string, make it an iterable list
                restart_file_dests = [restart_file_dests]

            if len(restart_file_dests) == 1:
                # if only one nenry, then all possible restart files go to that
                # destination
                restart_file_dests = restart_file_dests*len(
                    restart_file_glob_patterns)

            if len(restart_file_dests) > 1:
                # supposedly, specific destinations have been specified for
                # all possible restart files. If not:
                if len(restart_file_glob_patterns) != len(restart_file_dests):
                    logger.warning(
                        "There are {} restart_file_glob_patterns, "
                        "but {} restart_file_dests, latter ignored. "
                        "Specify none, a single or one "
                        "restart_file_dest per restart_file_glob_patterns "
                        "a Every restart file glob pattern ".format(
                            len(restart_file_glob_patterns),
                            len(restart_file_dests)))
                    # fall back to default
                    restart_file_dests = [os.curdir]*len(
                        restart_file_glob_patterns)

            # we have to decide whether the previous FireWorks failed or ended
            # successfully and then append a restart run or not

            recover = True  # per default, recover
            # check whether a previous firework handed down information
            prev_job_info = None
            path_prefix = None
            # pull from intentionally passed job info:
            if '_job_info' in fw_spec:
                job_info_array = fw_spec['_job_info']
                prev_job_info = job_info_array[-1]
                path_prefix = prev_job_info['launch_dir']
                logger.info('The name of the previous job was: {}'.format(
                    prev_job_info['name']))
                logger.info('The id of the previous job was: {}'.format(
                    prev_job_info['fw_id']))
                logger.info('The location of the previous job was: {}'.format(
                    path_prefix))
            # TODO: fails for several parents if the "wrong" parent fizzles
            # pull from fizzled previous FW:
            elif '_fizzled_parents' in fw_spec:
                fizzled_parents_array = fw_spec['_fizzled_parents']
                # pull latest (or last) fizzled parent:
                prev_job_info = fizzled_parents_array[-1]
                # pull latest launch
                path_prefix = prev_job_info['launches'][-1]['launch_dir']
                logger.info(
                    'The name of fizzled parent Firework was: {}'.format(
                        prev_job_info['name']))
                logger.info(
                    'The id of fizzled parent Firework was: {}'.format(
                        prev_job_info['fw_id']))
                logger.info(
                    'The location of fizzled parent Firework was: {}'.format(
                        path_prefix))
            else:  # no info about previous (fizzled or other) jobs
                logger.info(
                    'No information about previous (fizzled or other) jobs available.')
                recover = False  # don't recover
                # assume that parent completed successfully

            if prev_job_info and 'spec' in prev_job_info:
                logger.debug("Parent Firewor has 'spec':")
                _log_nested_dict(logger.debug, prev_job_info["spec"])
            else:
                logger.warning("Parent Firework description does not include 'spec' key.")

            # find other files to forward:
            file_list = []

            if recover:
                if not isinstance(other_glob_patterns, Iterable):
                    other_glob_patterns = [other_glob_patterns]
                for other_glob_pattern in other_glob_patterns:
                    if isinstance(other_glob_pattern, str):  # avoid non string objs
                        logger.info("Processing glob pattern {}".format(
                            other_glob_pattern))
                        file_list.extend(
                            glob.glob(
                                os.path.join(
                                    path_prefix, other_glob_pattern))
                        )

                # copy other files if necessary
                if len(file_list) > 0:
                    for f in file_list:
                        logger.info("File {} will be forwarded.".format(f))
                        try:
                            dest = os.getcwd()
                            shutil.copy(f, dest)
                        except Exception as exc:
                            if ignore_errors:
                                logger.warning("There was an error copying "
                                               "'{}' to '{}', ignored:".format(
                                                   f, dest))
                                logger.warning(exc)
                            else:
                                raise exc

                # find restart files as (src, dest) tuples:
                restart_file_list = []

                for glob_pattern, dest in zip(restart_file_glob_patterns,
                                              restart_file_dests):
                    restart_file_matches = glob.glob(os.path.join(
                        path_prefix, glob_pattern))

                    # determine most recent of restart files matches:
                    if len(restart_file_matches) > 1:
                        sorted_restart_file_matches = sorted(
                            restart_file_matches, key=os.path.getmtime)  # sort by modification time
                        logger.info("Several restart files {} (most recent last) "
                                    "for glob pattern '{}'.".format(
                                        glob_pattern,
                                        sorted_restart_file_matches))
                        logger.info("Modification times for those files: {}".format(
                            [os.path.getmtime(f) for f in sorted_restart_file_matches]))
                        logger.info("Most recent restart file '{}' will be copied "
                                    "to '{}'.".format(
                                        sorted_restart_file_matches[-1], dest))
                        restart_file_list.append(
                            (sorted_restart_file_matches[-1], dest))
                    elif len(restart_file_matches) == 1:
                        logger.info("One restart file '{}' for glob "
                                    "pattern '{}' will be copied to '{}'.".format(
                                        restart_file_matches[0],
                                        glob_pattern, dest))
                        restart_file_list.append(
                            (restart_file_matches[0], dest))
                    else:
                        logger.info("No restart file!")
                        if fizzle_on_no_restart_file:
                            raise ValueError(
                                "No restart file in {} for glob pattern {}".format(
                                    path_prefix, glob_pattern))

                # copy all identified restart files
                if len(restart_file_list) > 0:
                    for current_restart_file, dest in restart_file_list:
                        current_restart_file_basename = os.path.basename(current_restart_file)
                        logger.info("File {} will be forwarded.".format(
                            current_restart_file_basename))
                        try:
                            shutil.copy(current_restart_file, dest)
                        except Exception as exc:
                            logger.error("There was an error copying from {} "
                                         "to {}".format(
                                             current_restart_file, dest))
                            raise exc

            # distinguish between FireWorks and Workflows by top-level keys
            # fw: ['spec', 'fw_id', 'created_on', 'updated_on', 'name']
            # wf: ['fws', 'links', 'name', 'metadata', 'updated_on', 'created_on']
            detour_wf = None
            addition_wf = None
            mapped_detour_fws_root = []
            mapped_detour_fws_leaf = []
            mapped_addition_fws_root = []

            # if detour_fw given, append in any case:
            if isinstance(detour_wf_dict, dict):
                detour_wf_base_spec = None
                if superpose_detour_on_parent_fw_spec:
                    if prev_job_info and ("spec" in prev_job_info):
                        detour_wf_base_spec = prev_job_info["spec"]
                    else:
                        logger.warning("Superposition of detour_wf's "
                                       "fw_spec onto parent's "
                                       "fw_spec desired, but not parent"
                                       "fw_spec recovered.")
                detour_wf, detour_wf_fw_id_mapping = self.appendable_wf_from_dict(
                    detour_wf_dict, base_spec=detour_wf_base_spec,
                    exclusions=detour_fw_spec_to_exclude_dict)

                if detour_fws_root is None:  # default, as in core fireworks
                    mapped_detour_fws_root.extend(detour_wf.root_fw_ids)
                elif isinstance(detour_fws_root, (list, tuple)):
                    mapped_detour_fws_root.extend(
                        [detour_wf_fw_id_mapping[fw_id] for fw_id in detour_fws_root])
                else:  # isinstance(detour_fws_root, int)
                    mapped_detour_fws_root.append(detour_wf_fw_id_mapping[detour_fws_root])

                if detour_fws_leaf is None:  # default, as in core fireworks
                    mapped_detour_fws_leaf.extend(detour_wf.leaf_fw_ids)
                elif isinstance(detour_fws_leaf, (list, tuple)):
                    mapped_detour_fws_leaf.extend(
                        [detour_wf_fw_id_mapping[fw_id] for fw_id in detour_fws_leaf])
                else:  # isinstance(detour_fws_leaf, int)
                    mapped_detour_fws_leaf.append(detour_wf_fw_id_mapping[detour_fws_leaf])

                # only log if sepcific roots or leaves had been specified
                if detour_fws_root:
                    logger.debug("Mapped detour_wf root fw_ids {} onto newly created fw_ids {}".format(
                        detour_fws_root, mapped_detour_fws_root[-len(detour_fws_root):]))
                if detour_fws_leaf:
                    logger.debug("Mapped detour_wf leaf fw_ids {} onto newly created fw_ids {}".format(
                        detour_fws_leaf, mapped_detour_fws_leaf[-len(detour_fws_leaf)]))

            if detour_wf is not None:
                logger.debug(
                    "detour_wf:")
                _log_nested_dict(logger.debug, detour_wf.as_dict())

            # append restart fireworks if desired
            if recover:
                # try to derive number of restart from fizzled parent
                restart_count = None
                if prev_job_info and ('spec' in prev_job_info):
                    try:
                        restart_count = get_nested_dict_value(
                            prev_job_info['spec'], restart_counter)
                    except KeyError:
                        logger.warning("Found no restart count in fw_spec of "
                                       "fizzled parent at key '{}.'".format(
                                            restart_counter))

                # if none found, look in own fw_spec
                if restart_count is None:
                    try:
                        restart_count = get_nested_dict_value(
                            prev_job_info['spec'], restart_counter)
                    except KeyError:
                        logger.warning("Found no restart count in own fw_spec "
                                       "at key '{}.'".format(restart_counter))

                # if still none found, assume it's the 1st
                if restart_count is None:
                    restart_count = 1  # original run without restart_count is "0th"
                else:  # make sure above's queried value is an integer
                    restart_count = int(restart_count) + 1  # next restart now

                if restart_count < max_restarts + 1:
                    logger.info(
                        "This is #{:d} of at most {:d} restarts.".format(
                            restart_count, max_restarts))

                    restart_wf_base_spec = None
                    if superpose_restart_on_parent_fw_spec:
                        if prev_job_info and ("spec" in prev_job_info):
                            restart_wf_base_spec = prev_job_info["spec"]
                        else:
                            logger.warning("Superposition of restart_wf's "
                                           "fw_spec onto parent's "
                                           "fw_spec desired, but not parent"
                                           "fw_spec recovered.")
                    restart_wf, restart_wf_fw_id_mapping = self.appendable_wf_from_dict(
                        restart_wf_dict, base_spec=restart_wf_base_spec,
                        exclusions=restart_fw_spec_to_exclude_dict)

                    if restart_fws_root is None:  # default, as in core fireworks
                        mapped_detour_fws_root.extend(restart_wf.root_fw_ids)
                    elif isinstance(restart_fws_root, (list, tuple)):
                        mapped_detour_fws_root.extend(
                            [restart_wf_fw_id_mapping[fw_id] for fw_id in restart_fws_root])
                    else:  # isinstance(restart_fws_root, int)
                        mapped_detour_fws_root.append(restart_wf_fw_id_mapping[restart_fws_root])

                    if restart_fws_leaf is None:
                        mapped_detour_fws_leaf.extend(restart_wf.leaf_fw_ids)
                    elif isinstance(restart_fws_leaf, (list, tuple)):
                        mapped_detour_fws_leaf.extend(
                            [restart_wf_fw_id_mapping[fw_id] for fw_id in restart_fws_leaf])
                    else:  # isinstance(restart_fws_leaf, int)
                        mapped_detour_fws_leaf.append(restart_wf_fw_id_mapping[restart_fws_leaf])

                    if restart_fws_root:
                        logger.debug("Mapped restart_wf root fw_ids {} onto newly created fw_ids {}".format(
                            restart_fws_root, mapped_detour_fws_root[-len(restart_fws_root):]))
                    if restart_fws_leaf:
                        logger.debug("Mapped restart_wf leaf fw_ids {} onto newly created fw_ids {}".format(
                            restart_fws_leaf, mapped_detour_fws_leaf[-len(restart_fws_leaf)]))

                    # apply updates to fw_spec
                    for fw in restart_wf.fws:
                        logger.debug("Insert restart counter '{}': {} into fw_id {}: '{}'.".format(
                            restart_counter, restart_count, fw.fw_id, fw.name))
                        set_nested_dict_value(
                            fw.spec, restart_counter, restart_count)

                    logger.debug(
                        "restart_wf:")
                    _log_nested_dict(logger.debug, restart_wf.as_dict())

                    # repeatedly append copy of this recover task:
                    recover_ft = self
                    logger.debug("subsequent recover_fw's task recover_ft:")
                    _log_nested_dict(logger.debug, recover_ft.as_dict())

                    # repeated recovery firework inherits the following specs:
                    # recover_fw_spec = {key: fw_spec[key] for key in fw_spec
                    #                   if key not in fw_spec_to_exclude}
                    recover_fw_spec = dict_merge({}, fw_spec,
                                                 merge_exclusions=fw_spec_to_exclude_dict)
                    set_nested_dict_value(recover_fw_spec, restart_counter, restart_count)
                    logger.debug("Insert restart counter '{}': {} into recover_fw.spec.".format(
                                 restart_counter, restart_count))
                    logger.debug("Propagate fw_spec = {} to subsequent "
                                 "recover_fw.".format(recover_fw_spec))

                    # merge insertions
                    #
                    #  + - - - - - - - - - - - - - - - - - - -+
                    #  ' detour_wf                            '
                    #  '                                      '
                    #  ' +------------------+     +---------+ '
                    #  ' | detour_wf roots  | --> | leaf(s) | ' ------+
                    #  ' +------------------+     +---------+ '       |
                    #  '                                      '       |
                    #  + - - - - - - - - - - - - - - - - - - -+       |
                    #                                                 |
                    #                                                 |
                    #                                                 |
                    #  + - - - - - - - - - - - - - - - - - - -+       |
                    #  ' restart_wf                           '       |
                    #  '                                      '       v
                    #  ' +------------------+     +---------+ '     +----------+
                    #  ' | restart_wf roots | --> | leaf(s) | ' --> | recovery |
                    #  ' +------------------+     +---------+ '     +----------+
                    #
                    # into one workflow and make repeated recovery fireworks
                    # dependent on all leaf fireworks of detour and restart:

                    if restart_wf is not None and detour_wf is not None:
                        detour_wf.append_wf(restart_wf, [])
                    elif restart_wf is not None:  # and detour wf is None
                        detour_wf = restart_wf

                    recover_fw = Firework(
                        recover_ft,
                        spec=recover_fw_spec,  # inherit this Firework's spec
                        name=repeated_recover_fw_name,
                        fw_id=self.consecutive_fw_id)
                    self.consecutive_fw_id -= 1
                    logger.info("Create repeated recover Firework {} with "
                                "id {} and specs {}".format(recover_fw.name,
                                                            recover_fw.fw_id,
                                                            recover_fw.spec))

                    recover_wf = Workflow([recover_fw])

                    detour_wf.append_wf(recover_wf, mapped_detour_fws_leaf)

                    # now we don't need the true 'mapped_detour_fws_leaf' anymore, recover_wf becomes the leaf
                    mapped_detour_fws_leaf = recover_wf.leaf_fw_ids
                else:
                    logger.warning(
                        "Maximum number of {} restarts reached. "
                        "No further restart.".format(max_restarts))
            else:
                logger.warning("No restart Fireworks appended.")

            if detour_wf is not None:
                self.write_files_prev(detour_wf, fw_spec, root_fw_ids=mapped_detour_fws_root)
                logger.debug(
                    "Workflow([*detour_wf.fws, *restart_wf.fws, recover_fw]):")
                _log_nested_dict(logger.debug, detour_wf.as_dict())

            if isinstance(addition_wf_dict, dict):
                addition_wf_base_spec = None
                if superpose_addition_on_parent_fw_spec:
                    if prev_job_info and ("spec" in prev_job_info):
                        addition_wf_base_spec = prev_job_info["spec"]
                    else:
                        logger.warning("Superposition of addition_wf's "
                                       "fw_spec onto parent's "
                                       "fw_spec desired, but not parent"
                                       "fw_spec recovered.")
                addition_wf, addition_wf_fw_id_mapping = self.appendable_wf_from_dict(
                    addition_wf_dict, base_spec=addition_wf_base_spec,
                    exclusions=addition_fw_spec_to_exclude_dict)

                if addition_fws_root is None:
                    mapped_addition_fws_root.extend(addition_wf.root_fw_ids)
                elif isinstance(addition_fws_root, (list, tuple)):
                    mapped_addition_fws_root.extend(
                        [addition_wf_fw_id_mapping[fw_id] for fw_id in addition_fws_root])
                else:  # isinstance(addition_fws_root, int)
                    mapped_addition_fws_root.append(addition_wf_fw_id_mapping[addition_fws_root])

                if addition_fws_root:
                    logger.debug("Mapped addition_wf root fw_ids {} onto newly created fw_ids {}".format(
                        addition_fws_root, mapped_addition_fws_root[-len(addition_fws_root):]))

                self.write_files_prev(addition_wf, fw_spec, root_fw_ids=mapped_addition_fws_root)

        # end of ExitStack context
        output = {}
        if store_stdlog:
            stdlog_stream.flush()
            output['stdlog'] = stdlog_stream.getvalue()

        fw_action = FWAction()

        if stored_data:
            fw_action.stored_data = output

        if hasattr(fw_action, 'propagate') and propagate:
            fw_action.propagate = propagate

        if output_key:  # inject into fw_spec
            fw_action.mod_spec = [{dict_mod: {output_key: output}}]

        if addition_wf and apply_mod_spec_to_addition_wf:
            apply_mod_spec(addition_wf, fw_action, fw_ids=mapped_addition_fws_root)

        if detour_wf and apply_mod_spec_to_detour_wf:
            apply_mod_spec(detour_wf, fw_action, fw_ids=mapped_detour_fws_root)

        if addition_wf:  # for now, only one addition
            fw_action.additions = [addition_wf]
            fw_action.additions_root_fw_ids = [mapped_addition_fws_root]

        if detour_wf:
            # the leaf of a detour_wf is either another recover_fw or
            # the leaves of detour_wf_dict
            fw_action.detours = [detour_wf]
            fw_action.detours_root_fw_ids = [mapped_detour_fws_root]
            fw_action.detours_leaf_fw_ids = [mapped_detour_fws_leaf]

        return fw_action



# coding: utf-8
"""Test dtool lookup server queries integration."""

__author__ = 'Johannes Laurin Hoermann'
__copyright__ = 'Copyright 2020, IMTEK Simulation, University of Freiburg'
__email__ = 'johannes.hoermann@imtek.uni-freiburg.de, johannes.laurin@gmail.com'
__date__ = 'Jan 18, 2021'

import glob
import json
import logging
import pytest

from fireworks.utilities.dict_mods import apply_mod

from jlhfw.utils.logging import _log_nested_dict

from jlhfw.fireworks.user_objects.firetasks.dataflow_tasks import SearchDictTask
from jlhfw.fireworks.user_objects.firetasks.dtool_tasks import FetchItemTask
from jlhfw.fireworks.user_objects.firetasks.dtool_lookup_tasks import (
    DirectReadmeTask, DirectManifestTask)
from jlhfw.utils.dict import compare, _make_marker

# from test_dtool_tasks import _compare


@pytest.fixture
def dtool_smb_config(dtool_config):
    """Provide default dtool lookup config."""
    dtool_config.update({
        "DTOOL_SMB_SERVER_NAME_test-share": "localhost",
        "DTOOL_SMB_SERVER_PORT_test-share": 4445,
        "DTOOL_SMB_USERNAME_test-share": "guest",
        "DTOOL_SMB_PASSWORD_test-share": "a-guest-needs-no-password",
        "DTOOL_SMB_DOMAIN_test-share": "WORKGROUP",
        "DTOOL_SMB_SERVICE_NAME_test-share": "sambashare",
        "DTOOL_SMB_PATH_test-share": "dtool"
    })
    return dtool_config


@pytest.fixture
def default_direct_readme_dtool_task_spec(dtool_smb_config):
    """Provide default test task_spec for ReadmeDtoolTask."""
    return {
        'dtool_config': dtool_smb_config,
        'stored_data': True,
        'uri': 'smb://test-share/1a1f9fad-8589-413e-9602-5bbd66bfe675',
        'loglevel': logging.DEBUG,
    }


@pytest.fixture
def default_direct_manifest_dtool_task_spec(dtool_smb_config):
    """Provide default test task_spec for ManifestDtoolTask."""
    return {
        'dtool_config': dtool_smb_config,
        'stored_data': True,
        'uri': 'smb://test-share/1a1f9fad-8589-413e-9602-5bbd66bfe675',
        'loglevel': logging.DEBUG,
    }


#
# direct dtool lookup tasks tests
#


def test_direct_readme_dtool_task_run(default_direct_readme_dtool_task_spec):
    """Will lookup some dataset on the server."""
    logger = logging.getLogger(__name__)

    logger.debug("Instantiate DirectReadmeTask with '{}'".format(
        default_direct_readme_dtool_task_spec))

    t = DirectReadmeTask(**default_direct_readme_dtool_task_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())

    output = fw_action.stored_data['output']

    # TODO: dataset creation in test
    expected_respone = {
        "creation_date": "2020-11-08",
        "description": "testing description",
        "expiration_date": "2022-11-08",
        "funders": [
          {
            "code": "testing_code",
            "organization": "testing_organization",
            "program": "testing_program"
          }
        ],
        "owners": [
          {
            "email": "testing@test.edu",
            "name": "Testing User",
            "orcid": "testing_orcid",
            "username": "testing_user"
          }
        ],
        "project": "testing project"
    }

    assert compare(output, expected_respone)


def test_direct_manifest_dtool_task_run(default_direct_manifest_dtool_task_spec):
    """Will lookup some dataset on the server."""
    logger = logging.getLogger(__name__)

    logger.debug("Instantiate ManifestDtoolTask with '{}'".format(
        default_direct_manifest_dtool_task_spec))

    t = DirectManifestTask(**default_direct_manifest_dtool_task_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())

    output = fw_action.stored_data['output']

    # TODO: dataset creation in test
    expected_respone = {
        "dtoolcore_version": "3.18.0",
        "hash_function": "md5sum_hexdigest",
        "items": {
            "eb58eb70ebcddf630feeea28834f5256c207edfd": {
                "hash": "2f7d9c3e0cfd47e8fcab0c12447b2bf0",
                "relpath": "simple_text_file.txt",
                "size_in_bytes": 17,
                "utc_timestamp": 1606595093.53965
             }
        }
    }

    marker = {
        "dtoolcore_version": False,
        "hash_function": "md5sum_hexdigest",
        "items": {
            "eb58eb70ebcddf630feeea28834f5256c207edfd": {
                "hash": True,
                "relpath": True,
                "size_in_bytes": True,
                "utc_timestamp": False,
             }
        }
    }

    assert compare(output, expected_respone, marker)


# complex workflow

@pytest.fixture
def workflow_initial_fw_spec(dtool_smb_config):
    """Provide complex workflow test task_spec for QueryDtoolTask."""
    return {
        'initial_inputs': {
            'search': {
                'relpath': 'simple_text_file.txt',
            },
            'marker': {
                'relpath': True,
            },
        },
        'metadata': {
            "creation_date": "2020-11-28",
            "description": "description to override",
            "project": "derived testing project"
        }
    }


@pytest.fixture
def workflow_direct_readme_dtool_task_spec(dtool_smb_config):
    """Provide complex workflow test task_spec for QueryDtoolTask."""
    return {
        'dtool_config': dtool_smb_config,
        'stored_data': True,
        'uri': 'smb://test-share/1a1f9fad-8589-413e-9602-5bbd66bfe675',
        'output': 'metadata',
        'metadata_fw_source_key': 'metadata',
        'fw_supersedes_dtool': True,
        'metadata_dtool_exclusions': {'expiration_date': True},
        "metadata_fw_exclusions": {'description': True},
        'loglevel': logging.DEBUG
    }


@pytest.fixture
def workflow_direct_manifest_dtool_task_spec(dtool_smb_config):
    """Provide complex workflow test task_spec for ManifestDtoolTask."""
    return {
        'dtool_config': dtool_smb_config,
        'stored_data': True,
        'uri': 'smb://test-share/1a1f9fad-8589-413e-9602-5bbd66bfe675',
        'output': 'manifest_dtool_task->result',
        'loglevel': logging.DEBUG
    }


@pytest.fixture
def workflow_search_dict_task_spec():
    """Provide complex workflow test task_spec for SearchDictTask."""
    return {
        'input_key': 'manifest_dtool_task->result->items',
        'search_key': 'initial_inputs->search',
        'marker_key': 'initial_inputs->marker',
        'limit': 1,
        'expand': True,
        'stored_data': True,
        'output_key': 'search_dict_task->result',
        'loglevel': logging.DEBUG
    }


@pytest.fixture
def workflow_fetch_item_task_spec(dtool_smb_config):
    """Provide default test task_spec for CopyDatasetTask."""
    return {
        'item_id': {'key': 'search_dict_task->result'},
        'source': 'smb://test-share/1a1f9fad-8589-413e-9602-5bbd66bfe675',
        'filename': 'fetched_item.txt',
        'dtool_config': dtool_smb_config,
        'stored_data': True,
    }


def test_complex_direct_dtool_workflow(
        tempdir, workflow_initial_fw_spec,
        workflow_direct_readme_dtool_task_spec, workflow_direct_manifest_dtool_task_spec,
        workflow_search_dict_task_spec, workflow_fetch_item_task_spec,
        dtool_smb_config):
    """Query lookup server for datasets, subsequently query the readme, merge into fw_spec, then fetch specific file."""
    logger = logging.getLogger(__name__)

    # readme

    fw_spec = workflow_initial_fw_spec

    t = DirectReadmeTask(workflow_direct_readme_dtool_task_spec)
    logger.debug("Instantiated ReadmeDtoolTask as:")
    _log_nested_dict(logger.debug, t.as_dict())

    fw_action = t.run_task(fw_spec)
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())

    for mod in fw_action.mod_spec:
        apply_mod(mod, fw_spec)
    logger.debug("Modified fw_spec:")
    _log_nested_dict(logger.debug, fw_spec)
    
    expected_result = {
        "creation_date": "2020-11-28",  # from initial fw_spec
        "description": "testing description",  # from dtool readme
        # no "expiration_date" from dtool readme
        "funders": [
          {
            "code": "testing_code",
            "organization": "testing_organization",
            "program": "testing_program"
          }
        ],
        "owners": [
          {
            "email": "testing@test.edu",
            "name": "Testing User",
            "orcid": "testing_orcid",
            "username": "testing_user"
          }
        ],
        "project": "derived testing project",  # from fw_spec
    }
    marker = _make_marker(expected_result)

    assert compare(fw_spec['metadata'], expected_result, marker)

    # manifest

    t = DirectManifestTask(workflow_direct_manifest_dtool_task_spec)
    logger.debug("Instantiated ManifestDtoolTask as:")
    _log_nested_dict(logger.debug, t.as_dict())

    fw_action = t.run_task(fw_spec)
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())

    for mod in fw_action.mod_spec:
        apply_mod(mod, fw_spec)
    logger.debug("Modified fw_spec:")
    _log_nested_dict(logger.debug, fw_spec)

    expected_result = {
        "dtoolcore_version": "3.18.0",
        "hash_function": "md5sum_hexdigest",
        "items": {
            "eb58eb70ebcddf630feeea28834f5256c207edfd": {
                "hash": "2f7d9c3e0cfd47e8fcab0c12447b2bf0",
                "relpath": "simple_text_file.txt",
                "size_in_bytes": 17,
                "utc_timestamp": 1606595093.53965
             }
        }
    }

    marker = {
        "dtoolcore_version": False,
        "hash_function": True,
        "items": {
            "eb58eb70ebcddf630feeea28834f5256c207edfd": {
                "hash": True,
                "relpath": True,
                "size_in_bytes": True,
                "utc_timestamp": False,
             }
        }
    }

    assert compare(fw_spec['manifest_dtool_task']['result'], expected_result, marker)

    # search

    t = SearchDictTask(workflow_search_dict_task_spec)
    logger.debug("Instantiated SearchDictTask as:")
    _log_nested_dict(logger.debug, t.as_dict())

    fw_action = t.run_task(fw_spec)
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())

    for mod in fw_action.mod_spec:
        apply_mod(mod, fw_spec)
    logger.debug("Modified fw_spec:")
    _log_nested_dict(logger.debug, fw_spec)

    # fetch

    t = FetchItemTask(workflow_fetch_item_task_spec)
    fw_action = t.run_task(fw_spec)
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())

    local_item_files = glob.glob('*.txt')
    assert len(local_item_files)  == 1
    assert local_item_files[0] == 'fetched_item.txt'
    with open('fetched_item.txt', 'r') as f:
        content = f.read ()
    assert content == 'Some test content'




# coding: utf-8
"""Podman system service fixture."""

import json
import logging
import os
import pytest
import requests
import subprocess
import time

from fireworks.utilities.fw_serializers import ENCODING_PARAMS

from jlhfw.utils.logging import _log_nested_dict

from utils import _allocate_random_free_port

PODMAN_HOST = 'localhost'
PODMAN_RESTFUL_API_VERSION = '1.40.0'
PODMAN_RESTFUL_API_REQUEST_TEMPLATE = "http://{host:}:{port:}/v{version:}"


@pytest.fixture(scope="session")
def podman_system_service(request):
    """Launch rootless podman system service."""
    logger = logging.getLogger(__name__)

    podman_port = _allocate_random_free_port(PODMAN_HOST)

    protocol_host_port = "tcp:{host:}:{port:}".format(host=PODMAN_HOST, port=podman_port)

    cmd = ['podman', 'system', 'service', protocol_host_port,
           '--log-level', 'debug', '--timeout', '0'],

    logger.debug("Launching '{}'".format(' '.join(*cmd)))
    podman = subprocess.Popen(
        *cmd,
        stdin=None,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        env=os.environ, **ENCODING_PARAMS)

    # request = "http://localhost:8080/v1.40.0/libpod/info")
    request_template = PODMAN_RESTFUL_API_REQUEST_TEMPLATE.format(
        host=PODMAN_HOST, port=podman_port,
        version=PODMAN_RESTFUL_API_VERSION) + '/{}'
    req = request_template.format('libpod/info')
    while(True):
        try:
            response = requests.get(req)
        except requests.exceptions.ConnectionError as exc:
            logger.warning("Retrying after connection timeout: '{}'".format(exc))
            time.sleep(1)
        else:
            break

    if response.status_code != 200:
        logger.warning(
            "No connection via RESTful API to podman '{}: {}'".format(
                response.status_code, response.text))
        return
    info = json.loads(response.text)
    logger.info("Podman API info:")
    _log_nested_dict(logger.info, info)

    base_uri = PODMAN_RESTFUL_API_REQUEST_TEMPLATE.format(
        host=PODMAN_HOST, port=podman_port,
        version=PODMAN_RESTFUL_API_VERSION)

    def finalizer():
        """End podman system service."""
        ret = podman.poll()
        if isinstance(ret, int):
            logger.warning("podman system service returned with {}.".format(ret))
        else:  # None
            logger.info("Stopping podman system service....")
            podman.terminate()

        outs, errs = podman.communicate(timeout=3)

        logger.debug("podman system service stdout")
        for line in outs.splitlines():
            logger.debug(line)

        logger.debug("podman system service stderr")
        for line in errs.splitlines():
            logger.debug(line)

    request.addfinalizer(finalizer)
    return base_uri



import logging

from jlhfw.utils.logging import _log_nested_dict
from jlhfw.utils.dict import compare
from jlhfw.fireworks.user_objects.firetasks.dataflow_tasks import SearchDictTask

def test_search_dict_task_run():
    """Will lookup some dataset on the server."""
    logger = logging.getLogger(__name__)


    # TODO: dataset creation in test
    input = {
        "eb58eb70ebcddf630feeea28834f5256c207edfd": {
            "hash": "2f7d9c3e0cfd47e8fcab0c12447b2bf0",
            "relpath": "simple_text_file.txt",
            "size_in_bytes": 17,
            "utc_timestamp": 1606595093.53965
         },
        "eb58eb70ebcddf630feeea28834f5256c207edfe": {
            "hash": "2f7d9c3e0cfd47e8fcab0c12447b2bf1",
            "relpath": "another_simple_text_file.txt",
            "size_in_bytes": 17,
            "utc_timestamp": 1606595093.53965
        }
    }

    search = {"relpath": "simple_text_file.txt"}
    marker = {"relpath": True}

    task_spec = {
        'input': input,
        'search': search,
        'marker': marker,
        'stored_data': True,
    }

    t = SearchDictTask(**task_spec)
    fw_action = t.run_task({})

    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())

    output = fw_action.stored_data['output']

    expected_response = ['eb58eb70ebcddf630feeea28834f5256c207edfd']
    assert compare(output, expected_response)


"""smb share fixture."""

import logging
import pytest
import requests
import time
import urllib3

from smb.SMBConnection import SMBConnection
from smb.base import NotConnectedError

from utils import _allocate_random_free_port, _random_string

SAMBA_CONTAINER_IMAGE_NAME = 'docker.io/dperson/samba:latest'
SAMBA_HOST = 'localhost'

SAMBA_WORKGROUP = 'WORKGROUP'
SAMBA_SHARE = "sambashare"


@pytest.mark.usefixtures("podman_system_service")
@pytest.fixture
def smb_share(request, podman_system_service):
    """Provide smb share (via launching samba server container)."""
    logger = logging.getLogger(__name__)

    smb_port_139 = None
    smb_port_445 = None
    samba_container_name = None
    samba_container_id = None
    request_template = podman_system_service + '/{}'

    logger.info("Check image '{}' existance.".format(SAMBA_CONTAINER_IMAGE_NAME))
    uri_suffix = 'libpod/images/{}/exists'.format(SAMBA_CONTAINER_IMAGE_NAME)
    req = request_template.format(uri_suffix)
    response = requests.get(req)
    if response.status_code == 404:
        logger.info("Image '{}' does not exist, creation not implemented.".format(SAMBA_CONTAINER_IMAGE_NAME))
        # TODO: implement image creation
        return
    elif response.status_code != 204:
        logger.warning("Unexpected error at image existance check {}: '{}'.".format(
                       response.status_code, response.text))
        return None

    smb_port_139 = _allocate_random_free_port(SAMBA_HOST)
    smb_port_445 = _allocate_random_free_port(SAMBA_HOST)
    samba_container_name = _random_string()
    req = request_template.format('libpod/containers/create')
    request_params = {
        'image': SAMBA_CONTAINER_IMAGE_NAME,
        'name': samba_container_name,
        'portmappings': [
            {
                'container_port': 139,
                'host_port': smb_port_139,
            },
            {
                'container_port': 445,
                'host_port': smb_port_445,
            }
        ],
        'command': [
            "-p", "-S", "-w", "{workgroup}".format(workgroup=SAMBA_WORKGROUP),
            "-s", "{sambashare:};/share;yes;no;yes".format(sambashare=SAMBA_SHARE)
        ]
        # refer to https://hub.docker.com/r/dperson/samba
        # -s "<name;/path>[;browse;readonly;guest;users;admins;writelist;comment]"
        #             Configure a share
        #             required arg: "<name>;</path>"
        #             <name> is how it's called for clients
        #             <path> path to share
        #             NOTE: for the default values, just leave blank
        #             [browsable] default:'yes' or 'no'
        #             [readonly] default:'yes' or 'no'
        #             [guest] allowed default:'yes' or 'no'
        #             NOTE: for user lists below, usernames are separated by ','
        #             [users] allowed default:'all' or list of allowed users
        #             [admins] allowed default:'none' or list of admin users
        #             [writelist] list of users that can write to a RO share
        #             [comment] description of share
    }
    logger.info("Create container '{}': '{}'..".format(req, request_params))
    response = requests.post(req, json=request_params)
    if response.status_code == 201:
        samba_container_id = response.json()['Id']
        logger.info("Created container '{}': '{}'.".format(
                    samba_container_id, samba_container_name))
    else:
        logger.warning("Unexpected error at container creation {}: '{}'.".format(
                       response.status_code, response.text))
        return None

    uri_suffix = 'libpod/containers/{}/start'.format(samba_container_id)
    req = request_template.format(uri_suffix)
    logger.info("Start container '{}'".format(req))
    response = requests.post(req)
    if response.status_code == 204:
        logger.info("Container '{}': '{}' started successfully.".format(
            samba_container_id, samba_container_name))
    elif response.status_code == 304:
        logger.warning("Container '{}': '{}' already running.".format(
            samba_container_id, samba_container_name))
        # stop?
    else:
        logger.warning("Unexpected error at container launch {}: '{}'.".format(
                       response.status_code, response.text))
        ret = None

    conn = SMBConnection("guest", "a-guest-needs-no-password", "pytest",
                         SAMBA_HOST, use_ntlm_v2=True)

    while(True):
        try:
            conn_success = conn.connect(SAMBA_HOST, smb_port_445)
        except ConnectionResetError as exc:
            logger.warning("Connection to smb share failed ({}), retry...".format(exc))
            time.sleep(1)
        else:
            break

    if conn_success:
        ret = conn
        shares = conn.listShares()
        share_names = [s.name for s in shares]
        logger.info("Initial shares on smb host: {}.".format(share_names))
    else:
        logger.warning("Connection to smb share failed.")

    def finalizer():
        try:
            # TODO: check on connection state and reconnect if necessary
            shares = conn.listShares()
            share_names = [s.name for s in shares]
            logger.info("Eventual shares on smb host: {}.".format(share_names))

            for share in share_names:
                try:
                    content = conn.listPath(share, '')
                except Exception as exc:
                    logger.warning("Error on listing content of '{}': {}".format(share, exc))
                else:
                    content_names = [c.filename for c in content]
                    logger.info("Top-level content of share {}: {}.".format(share, content_names))

        except Exception as exc:
            # except NotConnectedError as exc:
            logger.warning("Error on listing files on smb share: {}".format(exc))

        try:
            conn.close()
        except Exception as exc:
            logger.warning("Error on closing smb connection: {}".format(exc))


        if samba_container_id:
            # stop container
            uri_suffix = "libpod/containers/{name:}/stop".format(name=samba_container_id)
            request = request_template.format(uri_suffix)
            try:
                response = requests.post(request)
            except urllib3.exceptions.MaxRetryError as exc:
                logger.warning("Unexpected exception '{}' when sending conatiner stop request.".format(exc))
            else:
                if response.status_code == 204:
                    logger.info("Container '{}': '{}' stopped successfully.".format(
                        samba_container_id, samba_container_name))
                elif response.status_code == 304:
                    logger.warning("Container '{}': '{}' not running.".format(
                        samba_container_id, samba_container_name))
                else:
                    logger.error(
                        "Unexpected error when stopping container '{}': '{}' - {}.".format(
                            samba_container_id, samba_container_name, response.text))

            # remove container
            uri_suffix = "libpod/containers/{name:}".format(name=samba_container_id)
            try:
                request = request_template.format(uri_suffix)
            except urllib3.exceptions.MaxRetryError as exc:
                logger.warning("Unexpected exception '{}' when sending conatiner stop request.".format(exc))
            else:
                response = requests.delete(request, params={'v': True})  # v: delete attached volumes
                if response.status_code == 204:
                    logger.info("Container '{}': '{}' removed successfully.".format(
                        samba_container_id, samba_container_name))
                else:
                    logger.error(
                        "Unexpected error when stopping container '{}': '{}' - {}.".format(
                            samba_container_id, samba_container_name, response.text))

    request.addfinalizer(finalizer)
    return ret



# coding: utf-8
"""Test dtool integration.

To see verbose logging during testing, run something like

    import logging
    import unittest
    from jlhfw.fireworks.user_objects.firetasks.tests.test_dtool_tasks import DtoolTasksTest
    logging.basicConfig(level=logging.DEBUG)
    suite = unittest.TestSuite()
    suite.addTest(DtoolTasksTest('name_of_the_desired_test'))
    runner = unittest.TextTestRunner()
    runner.run(suite)
"""
__author__ = 'Johannes Laurin Hoermann'
__copyright__ = 'Copyright 2020, IMTEK Simulation, University of Freiburg'
__email__ = 'johannes.hoermann@imtek.uni-freiburg.de, johannes.laurin@gmail.com'
__date__ = 'May 7, 2020'

import logging
import unittest
import os
import tempfile
import threading
import time
import mockssh

# needs dtool cli for verification

from jlhfw.fireworks.user_objects.firetasks.ssh_tasks import SSHForwardTask

module_dir = os.path.abspath(os.path.dirname(__file__))


class SSHTasksTest(unittest.TestCase):
    def setUp(self):
        # logger = logging.getLogger(__name__)

        self._tmpdir = tempfile.TemporaryDirectory()
        self._previous_working_directory = os.getcwd()
        os.chdir(self._tmpdir.name)

        self.default_ssh_forward_task_spec = {
            'remote_host': 'localhost',
            'remote_port': 80,
            'ssh_host': 'localhost',
            'ssh_user': 'testuser',
            'local_port': None,  # automatic allocation
            'port_file': '.port',
            # 'ssh_port': 22,
            'ssh_keyfile': os.path.join(module_dir, "id_rsa"),
        }

        self.default_ssh_user = {'testuser': os.path.join(module_dir, "id_rsa")}

    def tearDown(self):
        os.chdir(self._previous_working_directory)
        self._tmpdir.cleanup()

    def test_ssh_forward_task_run(self):
        """Establish ssh forwarding via mock ssh server."""
        logger = logging.getLogger(__name__)

        with mockssh.Server(self.default_ssh_user) as s:
            ssh_port = s.port
            logger.info("Started mock ssh server for users '{}' at port {}."
                .format(s.users, ssh_port))
            logger.debug("Instantiate SSHForwardTask with '{}'".format(
                {**self.default_ssh_forward_task_spec, 'ssh_port': ssh_port}))
            t = SSHForwardTask(ssh_port=ssh_port, **self.default_ssh_forward_task_spec)
            e = threading.Event()
            t.set_stop_event(e)
            thread = threading.Thread(target=t.run_task, kwargs={'fw_spec': {}})
            thread.start()
            logger.info("Started thread '{}.'".format(thread.name))
            time.sleep(5)
            logger.info("Send stop signal.")
            e.set()
            thread.join()
            logger.info("Thread ended.")


if __name__ == '__main__':
    unittest.main()



# coding: utf-8
"""Test dtool integration.

To see verbose logging during testing, run something like

    import logging
    import unittest
    from jlhfw.fireworks.user_objects.firetasks.tests.test_recover_tasks import RecoverTasksTest
    logging.basicConfig(level=logging.DEBUG)
    suite = unittest.TestSuite()
    suite.addTest(RecoverTasksTest('test_do_nothing'))
    runner = unittest.TextTestRunner()
    runner.run(suite)
"""
__author__ = 'Johannes Laurin Hoermann'
__copyright__ = 'Copyright 2020, IMTEK Simulation, University of Freiburg'
__email__ = 'johannes.hoermann@imtek.uni-freiburg.de, johannes.laurin@gmail.com'
__date__ = 'Aug 11, 2020'

import glob
import json
import logging
import os
import sys
import tempfile
import unittest

# needs dtool cli for verification
# import subprocess  # TODO: replace cli sub-processes with custom verify calls

from fireworks.core.firework import Firework, Workflow
from fireworks.user_objects.firetasks.script_task import PyTask
from jlhfw.fireworks.user_objects.firetasks.recover_tasks import RecoverTask, dict_merge

module_dir = os.path.abspath(os.path.dirname(__file__))


def _log_nested_dict(dct):
    logger = logging.getLogger(__name__)
    for l in json.dumps(dct, indent=2, default=str).splitlines():
        logger.debug(l)


def _read_json(file):
    with open(file, 'r') as stream:
        return json.load(stream)


# def _read_yaml(file):
#     with open(file, 'r') as stream:
#         return yaml.safe_load(stream)


def _inject(source, injection, marker):
    """Inserts values into nested dicts at positions marked by marker."""
    logger = logging.getLogger(__name__)
    if isinstance(marker, dict):
        for k, v in marker.items():
            if isinstance(v, str):  # str in marker marks desired mapping
                # inject here
                logger.debug(
                    "'{}' from injection[{}] overrides '{}' in source[{}]."
                        .format(injection[v], v, source[k], k))
                source[k] = injection[v]
            else:
                # descend
                logger.debug("Descending into sub-tree '{}' of '{}'.".format(
                    source[k], source))
                source[k] = _inject(source[k], injection, marker[k])
    elif isinstance(marker, list):  # source and marker must have same length
        logger.debug("Branching into element wise sub-trees of '{}'.".format(
            source))
        source = [_inject(s, injection, m) for s, m in zip(source, marker)]
    else:  # nothing to be done for this key
        logger.debug("No injection desired at '{}'.".format(source))

    return source


def _compare(source, target, marker):
    """Compare source and target partially, as marked by marker."""
    # ret = True
    logger = logging.getLogger(__name__)
    if isinstance(marker, dict):
        for k, v in marker.items():
            if k not in source:
                logger.error("{} not in source '{}'.".format(k, source))
                return False
            if k not in target:
                logger.error("{} not in target '{}'.".format(k, source))
                return False

            logger.debug("Descending into sub-tree '{}' of '{}'.".format(
                source[k], source))
            # descend
            if not _compare(source[k], target[k], v):
                return False  # one failed comparison suffices

    # source, target and marker must have same length:
    elif isinstance(marker, list):
        logger.debug("Branching into element wise sub-trees of '{}'.".format(
            source))
        for s, t, m in zip(source, target, marker):
            if not _compare(s, t, m):
                return False  # one failed comparison suffices
    else:  # arrived at leaf, comparison desired?
        if marker is not False:  # yes
            logger.debug("Comparing '{}' == '{}' -> {}.".format(
                source, target, source == target))
            return source == target

    # comparison either not desired or successfull for all elements
    return True


def _log_dir(directory=os.curdir):
    """Log content of directory subtree."""
    logger = logging.getLogger(__name__)
    for root, dirs, files in os.walk(directory, topdown=True):
        for name in files:
            logger.debug(os.path.join(root, name))
        for name in dirs:
            logger.debug(os.path.join(root, name))


def _get_dummy_restart_wf():
    # +-----------------+     +-----------------+     +------------------+
    # | restart_wf root | --> | restart_wf body | --> | restart_wf leafs |
    # +-----------------+     +-----------------+     +------------------+
    dummy_ft = PyTask(func='print', args=['I am a dummy task'])
    restart_root_fw = Firework([dummy_ft], name='restart_wf root')
    restart_body_fw = Firework([dummy_ft], name='restart_wf body', parents=[restart_root_fw])
    restart_leaf_fw = Firework([dummy_ft], name='restart_wf leaf', parents=[restart_body_fw])
    restart_wf = Workflow([restart_root_fw, restart_body_fw, restart_leaf_fw])
    return restart_wf


def _get_dummy_addition_wf():
    # +------------------+     +------------------+     +-------------------+
    # | addition_wf root | --> | addition_wf body | --> | addition_wf leafs |
    # +------------------+     +------------------+     +-------------------+
    dummy_ft = PyTask(func='print', args=['I am a dummy task'])
    addition_root_fw = Firework([dummy_ft], name='addition_wf root')
    addition_body_fw = Firework([dummy_ft], name='addition_wf body', parents=[addition_root_fw])
    addition_leaf_fw = Firework([dummy_ft], name='addition_wf leaf', parents=[addition_body_fw])
    addition_wf = Workflow([addition_root_fw, addition_body_fw, addition_leaf_fw])
    return addition_wf


def _get_dummy_detour_wf():
    # +----------------+     +----------------+     +-----------------+
    # | detour_wf root | --> | detour_wf body | --> | detour_wf leafs |
    # +----------------+     +----------------+     +-----------------+
    dummy_ft = PyTask(func='print', args=['I am a dummy task'])
    detour_root_fw = Firework([dummy_ft], name='detour_wf root')
    detour_body_fw = Firework([dummy_ft], name='detour_wf body', parents=[detour_root_fw])
    detour_leaf_fw = Firework([dummy_ft], name='detour_wf leaf', parents=[detour_body_fw])
    detour_wf = Workflow([detour_root_fw, detour_body_fw, detour_leaf_fw])
    return detour_wf


class RecoverTasksTest(unittest.TestCase):

    def _assert_restart_wf_topology(self, wf, root_fw_id):
        self.assertEqual(wf.id_fw[root_fw_id].name, 'restart_wf root')
        root_fw_children = wf.links[root_fw_id]

        self.assertEqual(len(root_fw_children), 1)
        body_fw_id = root_fw_children[0]
        self.assertEqual(wf.id_fw[body_fw_id].name, 'restart_wf body')

        self.assertEqual(len(wf.links[body_fw_id]), 1)
        leaf_fw_id = wf.links[body_fw_id][0]
        self.assertEqual(wf.id_fw[leaf_fw_id].name, 'restart_wf leaf')

        return leaf_fw_id

    def _assert_addition_wf_topology(self, wf, root_fw_id):
        self.assertEqual(wf.id_fw[root_fw_id].name, 'addition_wf root')
        root_fw_children = wf.links[root_fw_id]

        self.assertEqual(len(root_fw_children), 1)
        body_fw_id = root_fw_children[0]
        self.assertEqual(wf.id_fw[body_fw_id].name, 'addition_wf body')

        self.assertEqual(len(wf.links[body_fw_id]), 1)
        leaf_fw_id = wf.links[body_fw_id][0]
        self.assertEqual(wf.id_fw[leaf_fw_id].name, 'addition_wf leaf')

        return leaf_fw_id

    def _assert_detour_wf_topology(self, wf, root_fw_id):
        self.assertEqual(wf.id_fw[root_fw_id].name, 'detour_wf root')
        root_fw_children = wf.links[root_fw_id]

        self.assertEqual(len(root_fw_children), 1)
        body_fw_id = root_fw_children[0]
        self.assertEqual(wf.id_fw[body_fw_id].name, 'detour_wf body')

        self.assertEqual(len(wf.links[body_fw_id]), 1)
        leaf_fw_id = wf.links[body_fw_id][0]
        self.assertEqual(wf.id_fw[leaf_fw_id].name, 'detour_wf leaf')

        return leaf_fw_id

    def setUp(self):
        dummy_ft = PyTask(func='print', args=['I am a dummy task'])
        dummy_fw = Firework([dummy_ft], name='dummy_fw')
        dummy_wf = Workflow([dummy_fw])

        self.default_restart_wf_dict = dummy_wf.to_dict()

        self.default_task_spec = {
            'loglevel': logging.DEBUG,
            'store_stdlog': True,
            'stored_data': True,
        }

        self.default_fw_spec = {}

        self._tmpdir = tempfile.TemporaryDirectory()
        self._previous_working_directory = os.getcwd()

        os.chdir(self._tmpdir.name)

    def tearDown(self):
        os.chdir(self._previous_working_directory)
        self._tmpdir.cleanup()

    def test_ignore_missing_restart_file(self):
        """Will run a recovery task that does nothing."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        task_spec = {
            'restart_wf': self.default_restart_wf_dict,
            'fizzle_on_no_restart_file': False,
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_job_info': [{
                'launch_dir': os.curdir,
                'name': 'dummy parent',
                'fw_id': 1,
            }]
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)

        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())
        additions = fw_action.additions
        # detours = fw_action.detours

        self.assertEqual(additions, [])

    def test_increase_custom_restart_counter(self):
        """Will run a recovery task and check for an increased restart counter."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        task_spec = {
            'restart_wf': self.default_restart_wf_dict,
            'fizzle_on_no_restart_file': False,
            'restart_counter': 'nested->deeply_nested->restart_count',
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_fizzled_parents': [{
                'launch_dir': os.curdir,
                'name': 'dummy parent',
                'fw_id': 1,
                'spec': {
                    'nested': {'deeply_nested': {'restart_count': 2}},
                },
                'launches': [
                    {'launch_dir': './does/not/exist'}
                ]
            }]
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)

        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())

        detours = fw_action.detours[0]
        for fws in detours.fws:
            self.assertIn('nested',  fws.spec)
            self.assertIn('deeply_nested',  fws.spec["nested"])
            self.assertIn('restart_count',  fws.spec["nested"]["deeply_nested"])
            self.assertEqual(fws.spec["nested"]["deeply_nested"]["restart_count"], 3)

    def test_fizzle_on_no_restart_file(self):
        """Will run a recovery task that fizzles on no restart file."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        task_spec = {
            'restart_wf': self.default_restart_wf_dict,
            'fizzle_on_no_restart_file': True,
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_job_info': [{
                'launch_dir': os.curdir,
                'name': 'dummy parent',
                'fw_id': 1,
            }]
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)
        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))

        with self.assertRaises(ValueError) as cm:
            fw_action = t.run_task(fw_spec)

        self.assertEqual(str(cm.exception),
                         'No restart file in . for glob pattern *.restart[0-9]')


    def test_do_nothing_at_previous_success(self):
        """Will run a recovery task that does nothing."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        task_spec = {
            'restart_wf': self.default_restart_wf_dict,
            'fizzle_on_no_restart_file': False,
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {}
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)

        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())
        detours = fw_action.detours
        self.assertEqual(detours, [])

    def test_do_nothing_at_maximum_numnber_of_restarts(self):
        """Will run a recovery task that does nothing."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        task_spec = {
            'restart_wf': self.default_restart_wf_dict,
            'fizzle_on_no_restart_file': False,
            'max_restarts': 5,
            'restart_counter': 'restart_count',
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_fizzled_parents': [{
                'launches': [{'launch_dir': os.curdir}],
                'name': 'dummy parent',
                'fw_id': 1,
                'spec': {'restart_count': 5},
            }]
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)

        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())
        detours = fw_action.detours
        self.assertEqual(detours, [])

    def test_recover_single_restart_file(self):
        """Will run a recovery task that does nothing."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        os.mkdir('dummy_prev_launch')
        with open(os.path.join('dummy_prev_launch', 'dummy_restart_file'), 'w') as fp:
            pass

        logger.debug("Test directory tree before running RecoverTask.")
        _log_dir()

        task_spec = {
            'restart_wf': self.default_restart_wf_dict,
            'fizzle_on_no_restart_file': True,
            'restart_file_glob_patterns': '*_restart_file'
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_job_info': [{
                'launch_dir': 'dummy_prev_launch',
                'name': 'dummy parent',
                'fw_id': 1,
            }]
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)
        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())
        logger.debug("Test directory tree after running RecoverTask.")
        _log_dir()

        recoverd_restart_files = glob.glob('*_restart_file')
        self.assertEqual(len(recoverd_restart_files), 1)
        self.assertEqual(recoverd_restart_files[0], "dummy_restart_file")

    def test_select_latest_restart_file(self):
        """Will run a recovery task that does nothing."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        os.mkdir('dummy_prev_launch')
        first_restart_file = os.path.join('dummy_prev_launch', 'first_restart_file')
        second_restart_file = os.path.join('dummy_prev_launch', 'second_restart_file')
        with open(first_restart_file, 'w'):
            pass
        with open(second_restart_file, 'w'):
            pass

        logger.debug("Test directory tree before running RecoverTask.")
        _log_dir()

        # make sure files are sortable by mtime
        second_restart_file_stats = os.stat(second_restart_file)

        os.utime(first_restart_file, ns=(
            second_restart_file_stats.st_atime_ns-1e9,
            second_restart_file_stats.st_mtime_ns-1e9))

        first_restart_file_stats = os.stat(first_restart_file)

        logger.debug("firts_restart_file (atime_ns, mtime_ns) = ({}, {})".format(
                      first_restart_file_stats.st_atime_ns,
                      first_restart_file_stats.st_mtime_ns,))
        logger.debug("second_restart_file (atime_ns, mtime_ns) = ({}, {})".format(
                      second_restart_file_stats.st_atime_ns,
                      second_restart_file_stats.st_mtime_ns,))

        task_spec = {
            'restart_wf': self.default_restart_wf_dict,
            'fizzle_on_no_restart_file': True,
            'restart_file_glob_patterns': '*_restart_file'
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_job_info': [{
                'launch_dir': 'dummy_prev_launch',
                'name': 'dummy parent',
                'fw_id': 1,
            }]
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)
        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())
        logger.debug("Test directory tree after running RecoverTask.")
        _log_dir()

        recoverd_restart_files = glob.glob('*_restart_file')
        self.assertEqual(len(recoverd_restart_files), 1)
        self.assertEqual(recoverd_restart_files[0], "second_restart_file")

    def test_select_multiple_restart_files(self):
        """Will run a recovery task that does nothing."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        os.mkdir('dummy_prev_launch')

        first_restart_file = os.path.join('dummy_prev_launch', 'first_restart_file')
        second_restart_file = os.path.join('dummy_prev_launch', 'second_restart_file')
        with open(first_restart_file, 'w'):
            pass
        with open(second_restart_file, 'w'):
            pass

        logger.debug("Test directory tree before running RecoverTask.")
        _log_dir()

        # make sure files are sortable by mtime
        second_restart_file_stats = os.stat(second_restart_file)

        os.utime(first_restart_file, ns=(
            second_restart_file_stats.st_atime_ns-1e9,
            second_restart_file_stats.st_mtime_ns-1e9))

        first_restart_file_stats = os.stat(first_restart_file)

        logger.debug("firts_restart_file (atime_ns, mtime_ns) = ({}, {})".format(
                      first_restart_file_stats.st_atime_ns,
                      first_restart_file_stats.st_mtime_ns,))
        logger.debug("second_restart_file (atime_ns, mtime_ns) = ({}, {})".format(
                      second_restart_file_stats.st_atime_ns,
                      second_restart_file_stats.st_mtime_ns,))

        # different set of restart files
        first_checkpoint_file = os.path.join('dummy_prev_launch', 'first_checkpoint_file')
        second_checkpoint_file = os.path.join('dummy_prev_launch', 'second_checkpoint_file')
        with open(first_checkpoint_file, 'w'):
            pass
        with open(second_checkpoint_file, 'w'):
            pass

        logger.debug("Test directory tree before running RecoverTask.")
        _log_dir()

        # make sure files are sortable by mtime
        second_checkpoint_file_stats = os.stat(second_checkpoint_file)

        os.utime(first_checkpoint_file, ns=(
            second_checkpoint_file_stats.st_atime_ns-1e9,
            second_checkpoint_file_stats.st_mtime_ns-1e9))

        first_checkpoint_file_stats = os.stat(first_checkpoint_file)

        logger.debug("firts_checkpoint_file (atime_ns, mtime_ns) = ({}, {})".format(
                      first_checkpoint_file_stats.st_atime_ns,
                      first_checkpoint_file_stats.st_mtime_ns,))
        logger.debug("second_checkpoint_file (atime_ns, mtime_ns) = ({}, {})".format(
                      second_checkpoint_file_stats.st_atime_ns,
                      second_checkpoint_file_stats.st_mtime_ns,))

        task_spec = {
            'restart_wf': self.default_restart_wf_dict,
            'fizzle_on_no_restart_file': True,
            'restart_file_glob_patterns': ['*_restart_file', '*_checkpoint_file'],
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_job_info': [{
                'launch_dir': 'dummy_prev_launch',
                'name': 'dummy parent',
                'fw_id': 1,
            }]
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)
        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())
        logger.debug("Test directory tree after running RecoverTask.")
        _log_dir()

        recoverd_restart_files = glob.glob('*_restart_file')
        self.assertEqual(len(recoverd_restart_files), 1)
        self.assertEqual(recoverd_restart_files[0], "second_restart_file")

        recoverd_checkpoint_files = glob.glob('*_checkpoint_file')
        self.assertEqual(len(recoverd_checkpoint_files), 1)
        self.assertEqual(recoverd_checkpoint_files[0], "second_checkpoint_file")

    def test_recover_other_files(self):
        """Will run a recovery task that recovers all files matching other_glob_patterns."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        os.mkdir('dummy_prev_launch')

        some_file = os.path.join('dummy_prev_launch', 'some_file')
        some_other_file = os.path.join('dummy_prev_launch', 'some_other_file')
        some_third_file = os.path.join('dummy_prev_launch', 'some_third_file')
        with open(some_file, 'w'):
            pass
        with open(some_other_file, 'w'):
            pass
        with open(some_third_file, 'w'):
            pass

        logger.debug("Test directory tree before running RecoverTask.")
        _log_dir()

        task_spec = {
            'restart_wf': self.default_restart_wf_dict,
            'fizzle_on_no_restart_file': False,
            'other_glob_patterns': ['some_file', 'some_*_file'],
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_job_info': [{
                'launch_dir': 'dummy_prev_launch',
                'name': 'dummy parent',
                'fw_id': 1,
            }]
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)

        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())
        logger.debug("Test directory tree after running RecoverTask.")
        _log_dir()

        some_recovered_file = glob.glob('some_file')
        self.assertEqual(len(some_recovered_file), 1)
        self.assertEqual(some_recovered_file[0], "some_file")

        some_other_recovered_files = glob.glob('some_*_file')
        self.assertEqual(len(some_other_recovered_files), 2)
        self.assertIn("some_other_file", some_other_recovered_files)
        self.assertIn("some_third_file", some_other_recovered_files)

    def test_restart_wf_insertion(self):
        """Run a recovery task that returns a restart detour."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))
        # the following restart_wf parameter...
        # +-----------------+     +-----------------+     +------------------+
        # | restart_wf root | --> | restart_wf body | --> | restart_wf leafs |
        # +-----------------+     +-----------------+     +------------------+
        dummy_ft = PyTask(func='print', args=['I am a dummy task'])
        restart_root_fw = Firework([dummy_ft], name='restart_wf root')
        restart_body_fw = Firework([dummy_ft], name='restart_wf body', parents=[restart_root_fw])
        restart_leaf_fw = Firework([dummy_ft], name='restart_wf leaf', parents=[restart_body_fw])

        restart_wf = Workflow([restart_root_fw, restart_body_fw, restart_leaf_fw])

        logger.debug("Restart wf:")
        _log_nested_dict(restart_wf.as_dict())

        task_spec = {
            'fizzle_on_no_restart_file': False,
            'repeated_recover_fw_name': 'recover_fw',
            'restart_wf': restart_wf.as_dict(),
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_job_info': [{
                'launch_dir': 'dummy_prev_launch',
                'name': 'dummy parent',
                'fw_id': 1,
            }]
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)

        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())

        # ... must result in the following detour
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
        # ' restart_wf                                                          '
        # '                                                                     '
        # ' +-----------------+     +-----------------+     +-----------------+ '     +------------+
        # ' | restart_wf root | --> | restart_wf body | --> | restart_wf leaf | ' --> | recover_fw |
        # ' +-----------------+     +-----------------+     +-----------------+ '     +------------+
        # '                                                                     '
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
        detour = fw_action.detours[0]
        self.assertIsInstance(detour, Workflow)

        self.assertEqual(len(detour.leaf_fw_ids), 1)
        self.assertEqual(len(detour.root_fw_ids), 1)

        root_fw_id = detour.root_fw_ids[0]
        self.assertEqual(detour.id_fw[root_fw_id].name, 'restart_wf root')

        self.assertEqual(len(detour.links[root_fw_id]), 1)
        body_fw_id = detour.links[root_fw_id][0]
        self.assertEqual(detour.id_fw[body_fw_id].name, 'restart_wf body')

        self.assertEqual(len(detour.links[body_fw_id]), 1)
        leaf_fw_id = detour.links[body_fw_id][0]
        self.assertEqual(detour.id_fw[leaf_fw_id].name, 'restart_wf leaf')

        self.assertEqual(len(detour.links[leaf_fw_id]), 1)
        recover_fw_id = detour.links[leaf_fw_id][0]
        self.assertEqual(detour.id_fw[recover_fw_id].name, 'recover_fw')

    def test_detour_wf_insertion(self):
        """Run a recovery task that returns a detour."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        # The following restart_wf parameter...
        # +-----------------+     +-----------------+     +------------------+
        # | restart_wf root | --> | restart_wf body | --> | restart_wf leafs |
        # +-----------------+     +-----------------+     +------------------+
        dummy_ft = PyTask(func='print', args=['I am a dummy task'])
        restart_root_fw = Firework([dummy_ft], name='restart_wf root')
        restart_body_fw = Firework([dummy_ft], name='restart_wf body', parents=[restart_root_fw])
        restart_leaf_fw = Firework([dummy_ft], name='restart_wf leaf', parents=[restart_body_fw])

        restart_wf = Workflow([restart_root_fw, restart_body_fw, restart_leaf_fw])

        logger.debug("Restart wf:")
        _log_nested_dict(restart_wf.as_dict())

        # ...together with the following detour_wf parameter...
        # +----------------+     +-----------------+     +------------------+
        # | detour_wf root | --> | detour_wf body | --> | detour_wf leafs |
        # +----------------+     +-----------------+     +------------------+
        detour_root_fw = Firework([dummy_ft], name='detour_wf root')
        detour_body_fw = Firework([dummy_ft], name='detour_wf body', parents=[detour_root_fw])
        detour_leaf_fw = Firework([dummy_ft], name='detour_wf leaf', parents=[detour_body_fw])

        detour_wf = Workflow([detour_root_fw, detour_body_fw, detour_leaf_fw])

        logger.debug("Detour wf:")
        _log_nested_dict(detour_wf.as_dict())

        task_spec = {
            'restart_wf': restart_wf.as_dict(),
            'detour_wf': detour_wf.as_dict(),
            'fizzle_on_no_restart_file': False,
            'repeated_recover_fw_name': 'recover_fw',
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_job_info': [{
                'launch_dir': 'dummy_prev_launch',
                'name': 'dummy parent',
                'fw_id': 1,
            }]
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)

        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())

        # ... must result in the following detour:
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
        # ' detour_wf                                                           '
        # '                                                                     '
        # ' +-----------------+     +-----------------+     +-----------------+ '     +------------+
        # ' | detour_wf root  | --> | detour_wf body  | --> | detour_wf leaf  | ' --> | recover_fw |
        # ' +-----------------+     +-----------------+     +-----------------+ '     +------------+
        # '                                                                     '
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
        #                                                                               ^
        #                                                                               |
        #                                                                               |
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +       |
        # ' restart_wf                                                          '       |
        # '                                                                     '       |
        # ' +-----------------+     +-----------------+     +-----------------+ '       |
        # ' | restart_wf root | --> | restart_wf body | --> | restart_wf leaf | ' ------+
        # ' +-----------------+     +-----------------+     +-----------------+ '
        # '                                                                     '
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
        detour = fw_action.detours[0]
        self.assertIsInstance(detour, Workflow)

        self.assertEqual(len(detour.leaf_fw_ids), 1)
        self.assertEqual(len(detour.root_fw_ids), 2)

        # roots
        root_fw_names = {detour.id_fw[fw_id].name: fw_id for fw_id in detour.root_fw_ids}

        self.assertIn('detour_wf root', root_fw_names)
        self.assertIn('restart_wf root', root_fw_names)

        # restart_wf
        restart_root_fw_children = detour.links[root_fw_names['restart_wf root']]
        self.assertEqual(len(restart_root_fw_children), 1)
        restart_body_fw_id = restart_root_fw_children[0]
        self.assertEqual(detour.id_fw[restart_body_fw_id].name, 'restart_wf body')

        self.assertEqual(len(detour.links[restart_body_fw_id]), 1)
        restart_leaf_fw_id = detour.links[restart_body_fw_id][0]
        self.assertEqual(detour.id_fw[restart_leaf_fw_id].name, 'restart_wf leaf')

        self.assertEqual(len(detour.links[restart_leaf_fw_id]), 1)
        recover_fw_id = detour.links[restart_leaf_fw_id][0]
        self.assertEqual(detour.id_fw[recover_fw_id].name, 'recover_fw')

        # detour_wf
        detour_root_fw_children = detour.links[root_fw_names['detour_wf root']]
        self.assertEqual(len(detour_root_fw_children), 1)
        detour_body_fw_id = detour_root_fw_children[0]
        self.assertEqual(detour.id_fw[detour_body_fw_id].name, 'detour_wf body')

        self.assertEqual(len(detour.links[detour_body_fw_id]), 1)
        detour_leaf_fw_id = detour.links[detour_body_fw_id][0]
        self.assertEqual(detour.id_fw[detour_leaf_fw_id].name, 'detour_wf leaf')

        self.assertEqual(len(detour.links[detour_leaf_fw_id]), 1)
        same_recover_fw_id = detour.links[detour_leaf_fw_id][0]
        self.assertEqual(recover_fw_id, same_recover_fw_id)

    def test_addition_wf_insertion(self):
        """Run a recovery task that returns a detour."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        # The following restart_wf parameter...
        # +-----------------+     +-----------------+     +------------------+
        # | restart_wf root | --> | restart_wf body | --> | restart_wf leafs |
        # +-----------------+     +-----------------+     +------------------+
        dummy_ft = PyTask(func='print', args=['I am a dummy task'])
        restart_root_fw = Firework([dummy_ft], name='restart_wf root')
        restart_body_fw = Firework([dummy_ft], name='restart_wf body', parents=[restart_root_fw])
        restart_leaf_fw = Firework([dummy_ft], name='restart_wf leaf', parents=[restart_body_fw])

        restart_wf = Workflow([restart_root_fw, restart_body_fw, restart_leaf_fw])

        logger.debug("Restart wf:")
        _log_nested_dict(restart_wf.as_dict())

        # ...together with the following addition_wf parameter...
        # +----------------+     +-----------------+     +------------------+
        # | addition_wf root | --> | addition_wf body | --> | addition_wf leafs |
        # +----------------+     +-----------------+     +------------------+
        addition_root_fw = Firework([dummy_ft], name='addition_wf root')
        addition_body_fw = Firework([dummy_ft], name='addition_wf body', parents=[addition_root_fw])
        addition_leaf_fw = Firework([dummy_ft], name='addition_wf leaf', parents=[addition_body_fw])

        addition_wf = Workflow([addition_root_fw, addition_body_fw, addition_leaf_fw])

        logger.debug("Addition wf:")
        _log_nested_dict(addition_wf.as_dict())

        task_spec = {
            'restart_wf': restart_wf.as_dict(),
            'addition_wf': addition_wf.as_dict(),
            'fizzle_on_no_restart_file': False,
            'repeated_recover_fw_name': 'recover_fw',
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_job_info': [{
                'launch_dir': 'dummy_prev_launch',
                'name': 'dummy parent',
                'fw_id': 1,
            }]
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)

        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())

        # ... must result in the following addition:
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
        # ' addition_wf                                                            '
        # '                                                                        '
        # ' +------------------+     +------------------+     +------------------+ '
        # ' | addition_wf root | --> | addition_wf body | --> | addition_wf leaf | '
        # ' +------------------+     +------------------+     +------------------+ '
        # '                                                                        '
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
        #
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
        # ' restart_wf                                                          '
        # '                                                                     '
        # ' +-----------------+     +-----------------+     +-----------------+ '    +------------+
        # ' | restart_wf root | --> | restart_wf body | --> | restart_wf leaf | '--> | recover_fw |
        # ' +-----------------+     +-----------------+     +-----------------+ '    +------------+
        # '                                                                     '
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +

        detour = fw_action.detours[0]
        self.assertIsInstance(detour, Workflow)

        self.assertEqual(len(detour.leaf_fw_ids), 1)
        self.assertEqual(len(detour.root_fw_ids), 1)

        root_fw_id = detour.root_fw_ids[0]
        self.assertEqual(detour.id_fw[root_fw_id].name, 'restart_wf root')

        self.assertEqual(len(detour.links[root_fw_id]), 1)
        body_fw_id = detour.links[root_fw_id][0]
        self.assertEqual(detour.id_fw[body_fw_id].name, 'restart_wf body')

        self.assertEqual(len(detour.links[body_fw_id]), 1)
        leaf_fw_id = detour.links[body_fw_id][0]
        self.assertEqual(detour.id_fw[leaf_fw_id].name, 'restart_wf leaf')

        self.assertEqual(len(detour.links[leaf_fw_id]), 1)
        recover_fw_id = detour.links[leaf_fw_id][0]
        self.assertEqual(detour.id_fw[recover_fw_id].name, 'recover_fw')


        addition = fw_action.additions[0]
        self.assertIsInstance(addition, Workflow)

        self.assertEqual(len(addition.leaf_fw_ids), 1)
        self.assertEqual(len(addition.root_fw_ids), 1)

        root_fw_id = addition.root_fw_ids[0]
        self.assertEqual(addition.id_fw[root_fw_id].name, 'addition_wf root')

        self.assertEqual(len(addition.links[root_fw_id]), 1)
        body_fw_id = addition.links[root_fw_id][0]
        self.assertEqual(addition.id_fw[body_fw_id].name, 'addition_wf body')

        self.assertEqual(len(addition.links[body_fw_id]), 1)
        leaf_fw_id = addition.links[body_fw_id][0]
        self.assertEqual(addition.id_fw[leaf_fw_id].name, 'addition_wf leaf')

        self.assertEqual(len(addition.links[leaf_fw_id]), 0)


    def test_triple_wf_insertion(self):
        """Run a recovery task that returns a detour."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        restart_wf = _get_dummy_restart_wf()
        logger.debug("Restart wf:")
        _log_nested_dict(restart_wf.as_dict())


        addition_wf = _get_dummy_addition_wf()
        logger.debug("Addition wf:")
        _log_nested_dict(addition_wf.as_dict())

        detour_wf = _get_dummy_detour_wf()
        logger.debug("Detour wf:")
        _log_nested_dict(detour_wf.as_dict())


        task_spec = {
            'restart_wf': restart_wf.as_dict(),
            'addition_wf': addition_wf.as_dict(),
            'detour_wf': detour_wf.as_dict(),
            'fizzle_on_no_restart_file': False,
            'repeated_recover_fw_name': 'recover_fw',
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_job_info': [{
                'launch_dir': 'dummy_prev_launch',
                'name': 'dummy parent',
                'fw_id': 1,
            }]
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)

        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())

        # ... must result in the following addition:
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
        # ' addition_wf                                                            '
        # '                                                                        '
        # ' +------------------+     +------------------+     +------------------+ '
        # ' | addition_wf root | --> | addition_wf body | --> | addition_wf leaf | '
        # ' +------------------+     +------------------+     +------------------+ '
        # '                                                                        '
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
        #
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
        # ' detour_wf                                                           '
        # '                                                                     '
        # ' +-----------------+     +-----------------+     +-----------------+ '     +------------+
        # ' | detour_wf root  | --> | detour_wf body  | --> | detour_wf leaf  | ' --> | recover_fw |
        # ' +-----------------+     +-----------------+     +-----------------+ '     +------------+
        # '                                                                     '
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +
        #                                                                               ^
        #                                                                               |
        #                                                                               |
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +       |
        # ' restart_wf                                                          '       |
        # '                                                                     '       |
        # ' +-----------------+     +-----------------+     +-----------------+ '       |
        # ' | restart_wf root | --> | restart_wf body | --> | restart_wf leaf | ' ------+
        # ' +-----------------+     +-----------------+     +-----------------+ '
        # '                                                                     '
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - +

        # detour
        detour = fw_action.detours[0]
        self.assertIsInstance(detour, Workflow)

        self.assertEqual(len(detour.leaf_fw_ids), 1)
        self.assertEqual(len(detour.root_fw_ids), 2)

        # detour roots
        root_fw_names = {detour.id_fw[fw_id].name: fw_id for fw_id in detour.root_fw_ids}

        self.assertIn('detour_wf root', root_fw_names)
        self.assertIn('restart_wf root', root_fw_names)

        # restart_wf
        restart_leaf_fw_id = self._assert_restart_wf_topology(detour, root_fw_names['restart_wf root'])
        self.assertEqual(len(detour.links[restart_leaf_fw_id]), 1)
        recover_fw_id = detour.links[restart_leaf_fw_id][0]
        self.assertEqual(detour.id_fw[recover_fw_id].name, 'recover_fw')

        # detour_wf
        detour_leaf_fw_id = self._assert_detour_wf_topology(detour, root_fw_names['detour_wf root'])
        self.assertEqual(len(detour.links[detour_leaf_fw_id]), 1)
        same_recover_fw_id = detour.links[detour_leaf_fw_id][0]
        self.assertEqual(recover_fw_id, same_recover_fw_id)

        # addition
        addition = fw_action.additions[0]
        self.assertIsInstance(addition, Workflow)

        self.assertEqual(len(addition.leaf_fw_ids), 1)
        self.assertEqual(len(addition.root_fw_ids), 1)

        addition_root_fw_id = addition.root_fw_ids[0]
        addition_leaf_fw_id = self._assert_addition_wf_topology(addition, addition_root_fw_id)
        self.assertEqual(len(addition.links[addition_leaf_fw_id]), 0)

    def test_triple_wf_insertion_with_custom_links(self):
        """Run a recovery task that returns a detour."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        restart_wf = _get_dummy_restart_wf()
        logger.debug("Restart wf:")
        _log_nested_dict(restart_wf.as_dict())

        restart_wf_root_fw_ids = [fw.fw_id for fw in restart_wf.fws if fw.name == 'restart_wf body']
        restart_wf_leaf_fw_ids = [fw.fw_id for fw in restart_wf.fws if fw.name == 'restart_wf body']

        detour_wf = _get_dummy_detour_wf()
        logger.debug("Detour wf:")
        _log_nested_dict(detour_wf.as_dict())

        detour_wf_root_fw_ids = [fw.fw_id for fw in detour_wf.fws if fw.name == 'detour_wf body']
        detour_wf_leaf_fw_ids = [fw.fw_id for fw in detour_wf.fws if fw.name == 'detour_wf body']

        addition_wf = _get_dummy_addition_wf()
        logger.debug("Addition wf:")
        _log_nested_dict(addition_wf.as_dict())

        addition_wf_root_fw_ids = [fw.fw_id for fw in addition_wf.fws if fw.name == 'addition_wf body']

        task_spec = {
            'restart_wf': restart_wf.as_dict(),
            'addition_wf': addition_wf.as_dict(),
            'detour_wf': detour_wf.as_dict(),
            'restart_fws_root': restart_wf_root_fw_ids,
            'restart_fws_leaf': restart_wf_leaf_fw_ids,
            'detour_fws_root': detour_wf_root_fw_ids,
            'detour_fws_leaf': detour_wf_leaf_fw_ids,
            'addition_fws_root': addition_wf_root_fw_ids,
            'fizzle_on_no_restart_file': False,
            'repeated_recover_fw_name': 'recover_fw',
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_job_info': [{
                'launch_dir': 'dummy_prev_launch',
                'name': 'dummy parent',
                'fw_id': 1,
            }]
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)

        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())

        # per default, would result in
        #     + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
        #     ' restart_wf                                                             '
        #     '                                                                        '
        #     ' +------------------+     +------------------+     +------------------+ '     +------------+
        # --> ' | restart_wf root  | --> | restart_wf body  | --> | restart_wf leaf  | ' --> | recover_fw |
        #     ' +------------------+     +------------------+     +------------------+ '     +------------+
        #     '                                                                        '
        #     + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
        #                                                                                      ^
        #                                                                                      |
        #                                                                                      |
        #     + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+       |
        #     ' detour_wf                                                              '       |
        #     '                                                                        '       |
        #     ' +------------------+     +------------------+     +------------------+ '       |
        # --> ' |  detour_wf root  | --> |  detour_wf body  | --> |  detour_wf leaf  | ' ------+
        #     ' +------------------+     +------------------+     +------------------+ '
        #     '                                                                        '
        #     + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
        #     + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
        #     ' addition_wf                                                            '
        #     '                                                                        '
        #     ' +------------------+     +------------------+     +------------------+ '
        # --> ' | addition_wf root | --> | addition_wf body | --> | addition_wf leaf | '
        #     ' +------------------+     +------------------+     +------------------+ '
        #     '                                                                        '
        #     + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+

        # with custom entry and exit hooks (parent-child links), looks like
        #                              |
        #                              |
        #                              v
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
        # ' restart_wf                                                             '
        # '                                                                        '
        # ' +------------------+     +------------------+     +------------------+ '
        # ' | restart_wf root  | --> | restart_wf body  | --> | restart_wf leaf  | '
        # ' +------------------+     +------------------+     +------------------+ '
        # '                                                                        '
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
        #                              |
        #                              |
        #                              v
        #                            +------------------+
        #                            |    recover_fw    |
        #                            +------------------+
        #                              ^
        #                              |
        #                              |
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
        # ' detour_wf                                                              '
        # '                                                                        '
        # ' +------------------+     +------------------+     +------------------+ '
        # ' |  detour_wf root  | --> |  detour_wf body  | --> |  detour_wf leaf  | '
        # ' +------------------+     +------------------+     +------------------+ '
        # '                                                                        '
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
        #                              ^
        #                              |
        #                              |
        #
        #
        #
        #
        #
        #
        #                              |
        #                              |
        #                              v
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+
        # ' addition_wf                                                            '
        # '                                                                        '
        # ' +------------------+     +------------------+     +------------------+ '
        # ' | addition_wf root | --> | addition_wf body | --> | addition_wf leaf | '
        # ' +------------------+     +------------------+     +------------------+ '
        # '                                                                        '
        # + - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -+

        # detour
        detour = fw_action.detours[0]
        self.assertIsInstance(detour, Workflow)

        self.assertEqual(len(detour.leaf_fw_ids), 3)
        self.assertEqual(len(detour.root_fw_ids), 2)

        # detour roots
        root_fw_names = {detour.id_fw[fw_id].name: fw_id for fw_id in detour.root_fw_ids}
        self.assertIn('detour_wf root', root_fw_names)
        self.assertIn('restart_wf root', root_fw_names)

        # detour leaves
        leaf_fw_names = {detour.id_fw[fw_id].name: fw_id for fw_id in detour.leaf_fw_ids}
        self.assertIn('detour_wf leaf', leaf_fw_names)
        self.assertIn('restart_wf leaf', leaf_fw_names)
        self.assertIn('recover_fw', leaf_fw_names)

        recover_fw_id = leaf_fw_names['recover_fw']

        # restart_wf
        restart_root_fw_id = root_fw_names['restart_wf root']
        self.assertEqual(detour.id_fw[restart_root_fw_id].name, 'restart_wf root')
        restart_root_fw_children = detour.links[restart_root_fw_id]

        self.assertEqual(len(restart_root_fw_children), 1)
        restart_body_fw_id = restart_root_fw_children[0]
        self.assertEqual(detour.id_fw[restart_body_fw_id].name, 'restart_wf body')
        restart_body_fw_children = detour.links[restart_body_fw_id]

        self.assertEqual(len(restart_body_fw_children), 2)
        restart_body_fw_child_names = {detour.id_fw[fw_id].name: fw_id for fw_id in restart_body_fw_children}
        self.assertIn('restart_wf leaf', restart_body_fw_child_names)
        self.assertIn('recover_fw', restart_body_fw_child_names)
        self.assertEqual(restart_body_fw_child_names['recover_fw'], recover_fw_id)

        restart_leaf_fw_id = restart_body_fw_child_names['restart_wf leaf']
        self.assertEqual(len(detour.links[restart_leaf_fw_id]), 0)

        # detour_wf
        detour_root_fw_id = root_fw_names['detour_wf root']
        self.assertEqual(detour.id_fw[detour_root_fw_id].name, 'detour_wf root')
        detour_root_fw_children = detour.links[detour_root_fw_id]

        self.assertEqual(len(detour_root_fw_children), 1)
        detour_body_fw_id = detour_root_fw_children[0]
        self.assertEqual(detour.id_fw[detour_body_fw_id].name, 'detour_wf body')
        detour_body_fw_children = detour.links[detour_body_fw_id]

        self.assertEqual(len(detour_body_fw_children), 2)
        detour_body_fw_child_names = {detour.id_fw[fw_id].name: fw_id for fw_id in detour_body_fw_children}
        self.assertIn('detour_wf leaf', detour_body_fw_child_names)
        self.assertIn('recover_fw', detour_body_fw_child_names)
        self.assertEqual(detour_body_fw_child_names['recover_fw'], recover_fw_id)

        detour_leaf_fw_id = detour_body_fw_child_names['detour_wf leaf']
        self.assertEqual(len(detour.links[detour_leaf_fw_id]), 0)

        # addition
        addition = fw_action.additions[0]
        self.assertIsInstance(addition, Workflow)

        self.assertEqual(len(addition.leaf_fw_ids), 1)
        self.assertEqual(len(addition.root_fw_ids), 1)

        addition_root_fw_id = addition.root_fw_ids[0]
        self.assertEqual(addition.id_fw[addition_root_fw_id].name, 'addition_wf root')
        addition_root_fw_children = addition.links[addition_root_fw_id]

        self.assertEqual(len(addition_root_fw_children), 1)
        addition_body_fw_id = addition_root_fw_children[0]
        self.assertEqual(addition.id_fw[addition_body_fw_id].name, 'addition_wf body')

        self.assertEqual(len(addition.links[addition_body_fw_id]), 1)
        addition_leaf_fw_id = addition.links[addition_body_fw_id][0]
        self.assertEqual(addition.id_fw[addition_leaf_fw_id].name, 'addition_wf leaf')
        self.assertEqual(len(addition.links[addition_leaf_fw_id]), 0)

        # custom outgoing links
        self.assertEqual(len(fw_action.detours_root_fw_ids), 1)
        self.assertEqual(len(fw_action.detours_leaf_fw_ids), 1)
        self.assertEqual(len(fw_action.additions_root_fw_ids), 1)

        self.assertEqual(len(fw_action.detours_root_fw_ids[0]), 2)
        self.assertEqual(len(fw_action.detours_leaf_fw_ids[0]), 1)
        self.assertEqual(len(fw_action.additions_root_fw_ids[0]), 1)

        self.assertIn(restart_body_fw_id, fw_action.detours_root_fw_ids[0])
        self.assertIn(detour_body_fw_id, fw_action.detours_root_fw_ids[0])
        self.assertIn(recover_fw_id, fw_action.detours_leaf_fw_ids[0])
        self.assertIn(addition_body_fw_id, fw_action.additions_root_fw_ids[0])

        logger.info("detoors_root_fw_ids: {}".format(fw_action.detours_root_fw_ids))
        logger.info("detoors_leaf_fw_ids: {}".format(fw_action.detours_leaf_fw_ids))
        logger.info("additions_root_fw_ids: {}".format(fw_action.additions_root_fw_ids))

    def test_base_spec_for_triple_wf(self):
        """Run a recovery task that merges new and recovered specs for restart_wf, detour_wf, and restart_fw."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        restart_wf = _get_dummy_restart_wf()
        logger.debug("Restart wf:")
        _log_nested_dict(restart_wf.as_dict())


        addition_wf = _get_dummy_addition_wf()
        logger.debug("Addition wf:")
        _log_nested_dict(addition_wf.as_dict())

        detour_wf = _get_dummy_detour_wf()
        logger.debug("Detour wf:")
        _log_nested_dict(detour_wf.as_dict())


        task_spec = {
            'restart_wf': restart_wf.as_dict(),
            'addition_wf': addition_wf.as_dict(),
            'detour_wf': detour_wf.as_dict(),
            'fizzle_on_no_restart_file': False,
            'repeated_recover_fw_name': 'recover_fw',
            'fw_spec_to_exclude': {
                "top-level-key-to-include": {
                    "nested-level-key-to-exclude": True,
                },
                "top-level-key-to-exclude": True,
            },
            'restart_fw_spec_to_exclude': {
                "top-level-key-to-include": {
                    "nested-level-key-to-exclude": True,
                },
                "top-level-key-to-exclude": True,
            },
            'addition_fw_spec_to_exclude': {
                "top-level-key-to-include": {
                    "nested-level-key-to-exclude": True,
                },
                "top-level-key-to-exclude": True,
            },
            'detour_fw_spec_to_exclude': {
                "top-level-key-to-include": {
                    "nested-level-key-to-exclude": True,
                },
                "top-level-key-to-exclude": True,
            },
            "superpose_restart_on_parent_fw_spec": True,
            "superpose_addition_on_parent_fw_spec": True,
            "superpose_detour_on_parent_fw_spec": True,
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_job_info': [{
                'launch_dir': 'dummy_prev_launch',
                'name': 'dummy parent',
                'fw_id': 1,
                "spec": {
                    "top-level-key-to-include": {
                        "nested-level-key-to-include": True,
                        "nested-level-key-to-exclude": True,
                    },
                    "top-level-key-to-exclude": True
                },
            }],
            'top-level-key-in-recover-fw': True,
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)

        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())

        # ... must result in the following addition:

        # detour
        detour = fw_action.detours[0]
        self.assertIsInstance(detour, Workflow)

        self.assertEqual(len(detour.leaf_fw_ids), 1)
        self.assertEqual(len(detour.root_fw_ids), 2)

        # detour roots
        root_fw_names = {detour.id_fw[fw_id].name: fw_id for fw_id in detour.root_fw_ids}

        self.assertIn('detour_wf root', root_fw_names)
        self.assertIn('restart_wf root', root_fw_names)

        # restart_wf
        restart_leaf_fw_id = self._assert_restart_wf_topology(detour, root_fw_names['restart_wf root'])
        self.assertEqual(len(detour.links[restart_leaf_fw_id]), 1)
        recover_fw_id = detour.links[restart_leaf_fw_id][0]
        self.assertEqual(detour.id_fw[recover_fw_id].name, 'recover_fw')

        # detour_wf
        detour_leaf_fw_id = self._assert_detour_wf_topology(detour, root_fw_names['detour_wf root'])
        self.assertEqual(len(detour.links[detour_leaf_fw_id]), 1)
        same_recover_fw_id = detour.links[detour_leaf_fw_id][0]
        self.assertEqual(recover_fw_id, same_recover_fw_id)

        # addition
        addition = fw_action.additions[0]
        self.assertIsInstance(addition, Workflow)

        self.assertEqual(len(addition.leaf_fw_ids), 1)
        self.assertEqual(len(addition.root_fw_ids), 1)

        addition_root_fw_id = addition.root_fw_ids[0]
        addition_leaf_fw_id = self._assert_addition_wf_topology(addition, addition_root_fw_id)
        self.assertEqual(len(addition.links[addition_leaf_fw_id]), 0)

        # fw_spec assertions:
        for fw in [*detour.fws, *addition.fws]:
            if fw.name != "recover_fw":
                self.assertIn("top-level-key-to-include", fw.spec)
                self.assertIn("nested-level-key-to-include", fw.spec["top-level-key-to-include"])
                self.assertNotIn("nested-level-key-to-exclude", fw.spec["top-level-key-to-include"])
                self.assertNotIn("top-level-key-to-exclude", fw.spec)
                self.assertNotIn("top-level-key-in-recover-fw", fw.spec)
            else:
                self.assertNotIn("top-level-key-to-include", fw.spec)
                self.assertIn("top-level-key-in-recover-fw", fw.spec)

    def test_default_exclusions_for_triple_wf(self):
        """Run a recovery task that applies default exclusions when merging specs."""
        logger = logging.getLogger(__name__)
        logger.info("### {} ###".format(sys._getframe().f_code.co_name))

        restart_wf = _get_dummy_restart_wf()
        logger.debug("Restart wf:")
        _log_nested_dict(restart_wf.as_dict())


        addition_wf = _get_dummy_addition_wf()
        logger.debug("Addition wf:")
        _log_nested_dict(addition_wf.as_dict())

        detour_wf = _get_dummy_detour_wf()
        logger.debug("Detour wf:")
        _log_nested_dict(detour_wf.as_dict())


        task_spec = {
            'restart_wf': restart_wf.as_dict(),
            'addition_wf': addition_wf.as_dict(),
            'detour_wf': detour_wf.as_dict(),
            'fizzle_on_no_restart_file': False,
            'repeated_recover_fw_name': 'recover_fw',
            "superpose_restart_on_parent_fw_spec": True,
            "superpose_addition_on_parent_fw_spec": True,
            "superpose_detour_on_parent_fw_spec": True,
        }
        task_spec = dict_merge(self.default_task_spec, task_spec)

        fw_spec = {
            '_fizzled_parents': [{
                'name': 'dummy parent',
                'fw_id': 1,
                'launches': [
                    {'launch_dir': './does/not/exist'}
                ],
                'spec': {
                    '_fizzled_parents': {'from': 'FizzledParent'},
                    '_job_info': {'from': 'FizzledParent'},
                    '_fw_env': {'from': 'FizzledParent'},
                    '_files_prev': {'from': 'FizzledParent'},
                }
            }],
            '_fw_env': {'from': 'RecoverTask'},
            '_files_prev': {'from': 'RecoverTask'},
        }
        fw_spec = dict_merge(self.default_fw_spec, fw_spec)

        logger.debug("Instantiate RecoverTask with '{}'".format(task_spec))
        t = RecoverTask(**task_spec)
        logger.debug("Run with fw_spec '{}'".format(fw_spec))
        fw_action = t.run_task(fw_spec)
        logger.debug("FWAction:")
        _log_nested_dict(fw_action.as_dict())

        # ... must result in the following addition:

        # detour
        detour = fw_action.detours[0]
        addition = fw_action.additions[0]

        # fw_spec assertions:
        for fw in [*detour.fws, *addition.fws]:
            self.assertNotIn('_job_info', fw.spec)
            self.assertNotIn('_fw_env', fw.spec)
            self.assertNotIn('_files_prev', fw.spec)
            self.assertNotIn('_fizzled_parents', fw.spec)


if __name__ == '__main__':
    unittest.main()
